#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2015 Jérémy Bobbio <lunar@debian.org>
# Copyright © 2016 Ximin Luo <infinity0@debian.org>
# Copyright © 2016-2020 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import pytest
import subprocess

from diffoscope.config import Config
from diffoscope.comparators.ar import ArFile

from ..utils import diff_ignore_line_numbers
from ..utils.data import load_fixture, get_data
from ..utils.tools import (
    skip_unless_tools_exist,
    skip_unless_tool_is_at_least,
    skip_if_binutils_does_not_support_x86,
)
from ..utils.nonexisting import assert_non_existing
from ..utils.versions import Version


rlib1 = load_fixture("test1.rlib")
rlib2 = load_fixture("test2.rlib")


@pytest.fixture(scope="function", autouse=True)
def init_tests(request, monkeypatch):
    # Make sure decompilation is disabled so that tests don't break
    monkeypatch.setattr(Config(), "exclude_commands", ["^radare2.*"])


def llvm_version():
    return Version(
        subprocess.check_output(["llvm-config", "--version"])
        .decode("utf-8")
        .strip()
    )


def test_identification(rlib1):
    assert isinstance(rlib1, ArFile)


def test_no_differences(rlib1):
    difference = rlib1.compare(rlib1)
    assert difference is None


@pytest.fixture
def differences(rlib1, rlib2):
    return rlib1.compare(rlib2).details


@pytest.fixture
def rlib_dis_expected_diff():
    actual_ver = llvm_version()

    if actual_ver >= "3.8":
        diff_file = "rlib_llvm_dis_expected_diff"

    if actual_ver >= "5.0":
        diff_file = "rlib_llvm_dis_expected_diff_5"

    if actual_ver >= "7.0":
        diff_file = "rlib_llvm_dis_expected_diff_7"

    if actual_ver >= "10.0":
        diff_file = "rlib_llvm_dis_expected_diff_10"

    return get_data(diff_file)


@skip_unless_tools_exist("nm")
def test_num_items(differences):
    assert len(differences) == 4


@skip_unless_tools_exist("nm")
@skip_if_binutils_does_not_support_x86()
def test_item0_armap(differences):
    assert differences[0].source1 == "nm -s {}"
    assert differences[0].source2 == "nm -s {}"
    expected_diff = get_data("rlib_armap_expected_diff")
    assert differences[0].unified_diff == expected_diff


@skip_unless_tools_exist("nm")
@skip_if_binutils_does_not_support_x86()
def test_item1_elf(differences):
    assert differences[1].source1 == "alloc_system-d16b8f0e.0.o"
    assert differences[1].source2 == "alloc_system-d16b8f0e.0.o"
    expected_diff = get_data("rlib_elf_expected_diff")
    assert differences[1].details[0].unified_diff == expected_diff


@skip_unless_tools_exist("nm")
def test_item2_rust_metadata_bin(differences):
    assert differences[2].source1 == "rust.metadata.bin"
    assert differences[2].source2 == "rust.metadata.bin"


@skip_unless_tools_exist("llvm-dis")
@skip_unless_tool_is_at_least("llvm-config", llvm_version, "3.8")
def test_item3_deflate_llvm_bitcode(differences, rlib_dis_expected_diff):
    assert differences[3].source1 == "alloc_system-d16b8f0e.0.bytecode.deflate"
    assert differences[3].source2 == "alloc_system-d16b8f0e.0.bytecode.deflate"
    expected_diff = rlib_dis_expected_diff
    actual_diff = differences[3].details[0].details[1].unified_diff
    assert diff_ignore_line_numbers(actual_diff) == diff_ignore_line_numbers(
        expected_diff
    )


@skip_unless_tools_exist("nm")
def test_compare_non_existing(monkeypatch, rlib1):
    assert_non_existing(monkeypatch, rlib1)



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2017, 2020, 2022 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import os
import pytest

from diffoscope.comparators.binwalk import BinwalkFile

from ..utils.data import load_fixture, get_data
from ..utils.tools import skip_unless_tools_exist, skip_unless_module_exists
from ..utils.nonexisting import assert_non_existing

binwalk1 = load_fixture("test1.binwalk")
binwalk2 = load_fixture("test2.binwalk")

skip_if_root = pytest.mark.skipif(
    os.getuid() == 0, reason="cannot be run as root"
)


@skip_if_root
@skip_unless_module_exists("binwalk")
def test_identification(binwalk1, binwalk2):
    assert isinstance(binwalk1, BinwalkFile)
    assert isinstance(binwalk2, BinwalkFile)


@skip_if_root
@skip_unless_module_exists("binwalk")
def test_no_differences(binwalk1):
    difference = binwalk1.compare(binwalk1)
    assert difference is None


@pytest.fixture
def comparison(binwalk1, binwalk2):
    return binwalk1.compare(binwalk2)


@skip_if_root
@skip_unless_tools_exist("cpio")
@skip_unless_module_exists("binwalk")
def test_listing(comparison):
    differences = comparison.details
    assert comparison.comments == ["comprises of 2 embedded members"]
    assert differences[0].source1 == ".cpio file embedded at offset 0"
    assert differences[1].source2 == ".cpio file embedded at offset 600"

    expected_diff = get_data("binwalk_expected_diff")
    assert differences[0].details[0].unified_diff == expected_diff


@skip_if_root
@skip_unless_tools_exist("cpio")
@skip_unless_module_exists("binwalk")
def test_symlink(comparison):
    differences = comparison.details
    assert differences[0].details[1].source1 == "dir/link"
    assert differences[0].details[1].comment == "symlink"
    expected_diff = get_data("symlink_expected_diff")
    assert differences[0].details[1].unified_diff == expected_diff


@skip_if_root
@skip_unless_tools_exist("cpio")
@skip_unless_module_exists("binwalk")
def test_compare_non_existing(monkeypatch, binwalk1):
    assert_non_existing(monkeypatch, binwalk1)



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2018-2021 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import os
import glob
import subprocess

from .utils.tools import skip_unless_tool_is_at_least

ALLOWED_TEST_FILES = {
    # Data files we would prefer to generate dynamically
    "android1.img",
    "android2.img",
    "archive1.tar",
    "archive2.tar",
    "base-files_157-r45695_ar71xx.ipk",
    "base-files_157-r45918_ar71xx.ipk",
    "binary1",
    "binary2",
    "bug881937_1.deb",
    "bug881937_2.deb",
    "bug903391_1.deb",
    "bug903391_2.deb",
    "bug903401_1.deb",
    "bug903401_2.deb",
    "bug903565_1.deb",
    "bug903565_2.deb",
    "containers",
    "containers/a.tar.bz2",
    "containers/a.tar.gz",
    "containers/a.tar.xz",
    "containers/b.tar.bz2",
    "containers/b.tar.gz",
    "containers/b.tar.xz",
    "containers/magic_bzip2",
    "containers/magic_gzip",
    "containers/magic_xz",
    "dbgsym/add/test-dbgsym-dbgsym_1_amd64.deb",
    "dbgsym/add/test-dbgsym_1_amd64.deb",
    "dbgsym/mult/test-dbgsym-dbgsym_1_amd64.deb",
    "dbgsym/mult/test-dbgsym_1_amd64.deb",
    "dbgsym/test-dbgsym_1.dsc",
    "dbgsym/test-dbgsym_1.tar.gz",
    "debian-bug-876316-control.tar.gz",
    # Outputs
    "devicetree1.dtb",
    "devicetree2.dtb",
    "elfmix1.not_a",
    "elfmix2.a",
    "encrypted1.zip",
    "encrypted2.zip",
    "fuzzy-tar-in-tar1.tar",
    "fuzzy-tar-in-tar2.tar",
    "fuzzy1.tar",
    "fuzzy2.tar",
    "fuzzy3.tar",
    "hello1.wasm",
    "hello2.wasm",
    "no-perms.tar",
    "quine.gz",
    "quine.zip",
    "quine_a.deb",
    "quine_b.deb",
    "Samyak-Malayalam1.ttf",
    "Samyak-Malayalam2.ttf",
    "test1-le64.cache-4",
    "test1.a",
    "test1.apk",
    "test1.asc",
    "test1.binwalk",
    "test1.buildinfo",
    "test1.bz2",
    "test1.changes",
    "Test1.class",
    "Test2.class",
    "test1.cpio",
    "test1.db",
    "test1.deb",
    "test1.debsrc.tar.gz",
    "test1.debug",
    "test1.dex",
    "test1.docx",
    "test1.dsc",
    "test1.epub",
    "test1.exe",
    "test1.ext4",
    "test1.fat12",
    "test1.fat16",
    "test1.fat32",
    "test1.gif",
    "test1.git-index",
    "test1.gnumeric",
    "test1.gz",
    "test1.hi",
    "test1.icc",
    "test1.ico",
    "test1.inv",
    "test1.iso",
    "test1.jmod",
    "test1.jpg",
    "test1.js",
    "test1.json",
    "test1.kbx",
    "test1.lz4",
    "test1.macho",
    "test1.mo",
    "test1.mozzip",
    "test1.mp3",
    "test1.o",
    "test1.odt",
    "test1.ogg",
    "test1.pcap",
    "test1.pdf",
    "test1.pgp",
    "test1_signed.pgp",
    "test1.png",
    "test1.ppu",
    "test1.ps",
    "test1.pyc-renamed",
    "test1.rdb",
    "test1.rdx",
    "test1.rlib",
    "test1.rpm",
    "test1.sqlite3",
    "test1.squashfs",
    "test1_root.squashfs",
    "test1.tar",
    "test1.xml",
    "test1.xsb",
    "test1.xz",
    "test1.zip",
    "test1_meta.ico",
    "test1_meta.jpg",
    "test2-le64.cache-4",
    "test2.a",
    "test2.apk",
    "test2.asc",
    "test2.binwalk",
    "test2.buildinfo",
    "test2.bz2",
    "test2.changes",
    "test2.cpio",
    "test2.db",
    "test2.deb",
    "test2.debsrc.tar.gz",
    "test2.debug",
    "test2.dex",
    "test2.docx",
    "test2.dsc",
    "test2.epub",
    "test2.exe",
    "test2.ext4",
    "test2.fat12",
    "test2.gif",
    "test2.git-index",
    "test2.gnumeric",
    "test2.gz",
    "test2.hi",
    "test2.icc",
    "test2.ico",
    "test2.inv",
    "test2.iso",
    "test2.jmod",
    "test2.jpg",
    "test2.js",
    "test2.json",
    "test2.kbx",
    "test2.lz4",
    "test2.macho",
    "test2.mo",
    "test2.mozzip",
    "test2.mp3",
    "test2.o",
    "test2.odt",
    "test2.ogg",
    "test2.pcap",
    "test2.pdf",
    "test2.pgp",
    "test2_signed.pgp",
    "test2.png",
    "test2.ppu",
    "test2.ps",
    "test2.pyc-renamed",
    "test2.rdb",
    "test2.rdx",
    "test2.rlib",
    "test2.rpm",
    "test2.sqlite3",
    "test2.squashfs",
    "test2_root.squashfs",
    "test2.tar",
    "test2.xml",
    "test2.xsb",
    "test2.xz",
    "test2.zip",
    "test2_meta.ico",
    "test2_meta.jpg",
    "test3.apk",
    "test3.changes",
    "test3.gif",
    "test3.pdf",
    "test3.zip",
    "test4.changes",
    "test4.gif",
    "test4.pdf",
    "test_comment1.zip",
    "test_comment2.zip",
    "test_invalid.json",
    "test_invalid.xml",
    "test_iso8859-1.mo",
    "test_no_charset.mo",
    "test_openssh_pub_key1.pub",
    "test_openssh_pub_key2.pub",
    "test_weird_non_unicode_chars1.pdf",
    "test_weird_non_unicode_chars2.pdf",
    "text_ascii1",  # used by multiple tests
    "text_ascii2",  # used by multiple tests
    "text_iso8859",
    "text_order1",
    "text_order2",
    "text_unicode1",
    "text_unicode2",
    "text_unicode_binary_fallback",
    # Outputs
    "debian-bug-875281.collapsed-diff.json",
    "order1a.json",
    "order1b.json",
}


def black_version():
    try:
        out = subprocess.check_output(("black", "--version"))
    except subprocess.CalledProcessError as e:
        out = e.output

    # black --version format changed starting in 21.11b0. Returning the first
    # token starting with a decimal digit, since its ordinal position may vary.
    return [
        x for x in out.strip().decode("utf-8").split(" ") if x[0].isdigit()
    ]


@skip_unless_tool_is_at_least("black", black_version, "21.12b0")
def test_code_is_black_clean():
    output = subprocess.check_output(
        ("black", "--diff", "."), stderr=subprocess.PIPE
    ).decode("utf-8")

    # Display diff in "captured stdout call"
    if output:
        print(output)

    assert not output, output


def test_does_not_add_new_test_files():
    """
    For a variety of reasons we are now prefering to generate any test data
    dynamically (via pytest fixtures, etc.) rather than committing and shipping
    such files.

    Exceptions to this may be appropriate (or even required) but this test
    ensures that test files that could be dynamically generated are not added
    "automatically", for example by following previous/older commits.
    """

    test_dir = os.path.join(os.path.dirname(__file__), "data")

    seen = set()

    for x in glob.iglob(os.path.join(test_dir, "**"), recursive=True):
        if os.path.isdir(x):
            continue

        # Strip off common prefix
        x = x[len(test_dir) + 1 :]

        # Skip some known expected diff filename patterns
        if (
            x.endswith("_diff")
            or x.endswith("_diffs")
            or x.endswith(".diff")
            or "_diff_" in x
            or "diff." in x
            or x.startswith("output")
        ):
            continue

        seen.add(x)

    assert seen - ALLOWED_TEST_FILES - {""} == set()



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2016-2018, 2020 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import pytest

from diffoscope.tools import get_current_os
from diffoscope.comparators.git import GitIndexFile

from ..utils.data import assert_diff, load_fixture


git1 = load_fixture("test1.git-index")
git2 = load_fixture("test2.git-index")


def test_identification(git1):
    assert isinstance(git1, GitIndexFile)


def test_no_differences(git1):
    assert git1.compare(git1) is None


@pytest.fixture
def differences(git1, git2):
    return git1.compare(git2).details


def test_diff(differences):
    if get_current_os() == "FreeBSD":
        expected_diff = "git_expected_diff_freebsd"
    else:
        expected_diff = "git_expected_diff"
    assert_diff(differences[0], expected_diff)



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2018, 2020 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import pytest

from diffoscope.comparators.gnumeric import GnumericFile

from ..utils.data import load_fixture, assert_diff
from ..utils.tools import skip_unless_tools_exist
from ..utils.nonexisting import assert_non_existing

gnumeric1 = load_fixture("test1.gnumeric")
gnumeric2 = load_fixture("test2.gnumeric")


def test_identification(gnumeric1):
    assert isinstance(gnumeric1, GnumericFile)


def test_no_differences(gnumeric1):
    difference = gnumeric1.compare(gnumeric1)
    assert difference is None


@pytest.fixture
def differences(gnumeric1, gnumeric2):
    return gnumeric1.compare(gnumeric2).details


@skip_unless_tools_exist("ssconvert")
def test_diff(differences):
    assert_diff(differences[0], "gnumeric_expected_diff")


@skip_unless_tools_exist("ssconvert")
def test_compare_non_existing(monkeypatch, gnumeric1):
    assert_non_existing(monkeypatch, gnumeric1, has_null_source=False)



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2015 Jérémy Bobbio <lunar@debian.org>
# Copyright © 2016-2017, 2020 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import pytest

from diffoscope.config import Config
from diffoscope.comparators.zip import ZipFile
from diffoscope.comparators.missing_file import MissingFile

from ..utils.data import load_fixture, get_data
from ..utils.tools import skip_unless_tools_exist


epub1 = load_fixture("test1.epub")
epub2 = load_fixture("test2.epub")


def test_identification(epub1):
    assert isinstance(epub1, ZipFile)


def test_no_differences(epub1):
    difference = epub1.compare(epub1)
    assert difference is None


@pytest.fixture
def differences(epub1, epub2):
    return epub1.compare(epub2).details


@skip_unless_tools_exist("zipinfo")
def test_differences(differences):
    assert differences[0].source1 == "zipinfo {}"
    assert differences[0].source2 == "zipinfo {}"
    assert differences[1].source1 == "content.opf"
    assert differences[1].source2 == "content.opf"
    assert differences[2].source1 == "toc.ncx"
    assert differences[2].source2 == "toc.ncx"
    assert differences[3].source1 == "ch001.xhtml"
    assert differences[3].source2 == "ch001.xhtml"

    # Flatten everything recursively, as XMLFile will contain reformatted data
    # under Difference.details.
    def fn(difference):
        if difference.unified_diff:
            yield difference.unified_diff
        for x in difference.details:
            yield from fn(x)

    val = "\n".join("\n".join(fn(x)) for x in differences)

    assert val == get_data("epub_expected_diffs")


@skip_unless_tools_exist("zipinfo")
def test_compare_non_existing(monkeypatch, epub1):
    monkeypatch.setattr(Config(), "new_file", True)
    difference = epub1.compare(MissingFile("/nonexisting", epub1))
    assert difference.source2 == "/nonexisting"
    assert difference.details[-1].source2 == "/dev/null"



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2015 Daniel Kahn Gillmor <dkg@fifthhorseman.net>
# Copyright © 2015 Paul Gevers <elbrus@debian.org>
# Copyright © 2016-2018, 2020 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import pytest
import subprocess

from diffoscope.comparators.ppu import PpuFile

from ..utils.data import load_fixture, assert_diff
from ..utils.tools import skip_unless_tools_exist, skip_unless_tool_is_at_least
from ..utils.nonexisting import assert_non_existing


# These test files were taken from two different builds of the Debian package
# fp-units-castle-game-engine (version 5.1.1-2 on amd64) on the Debian
# reproducible build infrastructure. The files were originally called
# castletexturefont_dejavusans_10.ppu which are generated during package
# building of the cge package from dejavusans font in the fonts-dejavu package.

file1 = load_fixture("test1.ppu")
file2 = load_fixture("test2.ppu")


def ppudump_version():
    # first line of `PPU-Analyser Version 3.0.0` looks like:
    #   PPU-Analyser Version 3.0.0
    out = subprocess.check_output(["ppudump", "-h"])
    return out.decode("utf-8").splitlines()[0].split()[2].strip()


@skip_unless_tool_is_at_least("ppudump", ppudump_version, "3.2.0")
def test_identification(file1):
    assert isinstance(file1, PpuFile)


def test_no_differences(file1):
    assert file1.compare(file1) is None


@pytest.fixture
def differences(file1, file2):
    return file1.compare(file2).details


@skip_unless_tool_is_at_least("ppudump", ppudump_version, "3.2.0")
def test_diff(differences):
    assert_diff(differences[0], "ppu_expected_diff")


@skip_unless_tool_is_at_least("ppudump", ppudump_version, "3.2.0")
def test_compare_non_existing(monkeypatch, file1):
    assert_non_existing(monkeypatch, file1, has_null_source=False)



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2016 Reiner Herrmann <reiner@reiner-h.de>
# Copyright © 2016-2017, 2019-2020 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import pytest
import subprocess

from diffoscope.comparators.ps import PsFile

from ..utils.data import load_fixture, get_data
from ..utils.tools import skip_unless_tools_exist, skip_unless_tool_is_at_least
from ..utils.nonexisting import assert_non_existing


ps1 = load_fixture("test1.ps")
ps2 = load_fixture("test2.ps")


def ps2ascii_version():
    return subprocess.check_output(("ps2ascii", "--version")).decode("utf-8")


def test_identification(ps1):
    assert isinstance(ps1, PsFile)


@skip_unless_tool_is_at_least("ps2ascii", ps2ascii_version, "9.21")
def test_no_differences(ps1):
    difference = ps1.compare(ps1)
    assert difference is None


@pytest.fixture
def differences(ps1, ps2):
    return ps1.compare(ps2)


@skip_unless_tool_is_at_least("ps2ascii", ps2ascii_version, "9.21")
def test_internal_diff(differences):
    expected_diff = get_data("ps_internal_expected_diff")
    assert differences.unified_diff == expected_diff


@skip_unless_tool_is_at_least("ps2ascii", ps2ascii_version, "9.26")
def test_text_diff(differences):
    computed_diff = differences.details[0].unified_diff.replace("\r\n", "\n")
    expected_diff = get_data("ps_text_expected_diff")
    assert computed_diff == expected_diff


@skip_unless_tools_exist("ps2ascii")
@skip_unless_tool_is_at_least("ps2ascii", ps2ascii_version, "9.21")
def test_compare_non_existing(monkeypatch, ps1):
    assert_non_existing(monkeypatch, ps1, has_null_source=False)



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2016 Jérémy Bobbio <lunar@debian.org>
# Copyright © 2016-2020 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import pytest
import subprocess

from diffoscope.config import Config
from diffoscope.comparators.icc import IccFile
from diffoscope.comparators.missing_file import MissingFile

from ..utils.data import load_fixture, get_data, data
from ..utils.tools import skip_unless_tools_exist, skip_unless_tool_is_at_least


icc1 = load_fixture("test1.icc")
icc2 = load_fixture("test2.icc")


def cd_iccdump_version():
    """
    The `cd-iccdump` command has no --version command and, whilst the `colord`
    binary does, it relies on the daemon running to return any output.
    Therefore we hackily compare the output (via the line length) of colord
    1.4.3:

        "  Profile ID    = 0477fa4bb5ae5ae9a778f5cd72eb45a4"

    ... versus, for example, colord 1.3.3:

        " Profile ID    = 0x0477fa4b"

    We don't massage the output (say, in an `Iccdump.filter` method) as it
    would remove the accuracy and, unfortunately, colord 1.4.3 also removes a
    somewhat-arbitrary newline too.
    """

    val = subprocess.check_output(("cd-iccdump", data("test1.icc"))).decode(
        "utf-8"
    )

    for x in val.splitlines():
        if x.startswith("  Profile ID") and len(x) == 47:
            return "1.4.3"
    return "1.3.3"


def test_identification(icc1):
    assert isinstance(icc1, IccFile)


def test_no_differences(icc1):
    difference = icc1.compare(icc1)
    assert difference is None


@pytest.fixture
def differences(icc1, icc2):
    return icc1.compare(icc2).details


@skip_unless_tool_is_at_least("cd-iccdump", cd_iccdump_version, "1.4.3")
def test_diff(differences):
    if "ne_SU" in differences[0].unified_diff:
        pytest.skip(
            "Endian-specific differences detected; see "
            "<https://bugs.debian.org/847595>"
        )

    expected_diff = get_data("icc_expected_diff")
    assert differences[0].unified_diff == expected_diff


@skip_unless_tools_exist("cd-iccdump")
def test_compare_non_existing(monkeypatch, icc1):
    monkeypatch.setattr(Config(), "new_file", True)
    difference = icc1.compare(MissingFile("/nonexisting", icc1))
    assert difference.source2 == "/nonexisting"
    assert len(difference.details) > 0



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2016 Emanuel Bronshtein <e3amn2l@gmx.com>
# Copyright © 2016-2017, 2020 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import pytest

from diffoscope.config import Config
from diffoscope.comparators.javascript import JavaScriptFile
from diffoscope.comparators.missing_file import MissingFile

from ..utils.data import load_fixture, get_data
from ..utils.tools import skip_unless_tools_exist


javascript1 = load_fixture("test1.js")
javascript2 = load_fixture("test2.js")


def test_identification(javascript1):
    assert isinstance(javascript1, JavaScriptFile)


def test_no_differences(javascript1):
    difference = javascript1.compare(javascript1)
    assert difference is None


@pytest.fixture
def differences(javascript1, javascript2):
    return javascript1.compare(javascript2).details


@skip_unless_tools_exist("js-beautify")
def test_diff(differences):
    expected_diff = get_data("javascript_expected_diff")
    assert differences[0].unified_diff == expected_diff


@skip_unless_tools_exist("js-beautify")
def test_compare_non_existing(monkeypatch, javascript1):
    monkeypatch.setattr(Config(), "new_file", True)
    difference = javascript1.compare(MissingFile("/nonexisting", javascript1))
    assert difference.source2 == "/nonexisting"
    assert len(difference.details) > 0



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2017, 2019-2020 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

from .utils import Presenter


class RestructuredTextPresenter(Presenter):
    TITLE_CHARS = "=-`:.'\"~^_*+#"

    def __init__(self, print_func):
        self.print_func = print_func
        super().__init__()

    def visit_difference(self, difference):
        if difference.source1 == difference.source2:
            self.title(difference.source1)
        else:
            self.title(
                "Comparing {} & {}".format(
                    difference.source1, difference.source2
                )
            )

        for x in difference.comments:
            self.print_func()
            self.print_func(x)

        if difference.unified_diff:
            self.print_func("::")
            self.print_func()
            self.print_func(self.indent(difference.unified_diff, "    "))
            self.print_func()

    def title(self, val):
        char = self.TITLE_CHARS[self.depth % len(self.TITLE_CHARS)]

        if self.depth < len(self.TITLE_CHARS):
            self.print_func(len(val) * char)

        self.print_func(val)
        self.print_func(len(val) * char)
        self.print_func()



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2016 Brett Smith <debbug@brettcsmith.org>
# Copyright © 2016-2020 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import pytest

from diffoscope.path import set_path
from diffoscope.environ import normalize_environment
from diffoscope.progress import ProgressManager
from diffoscope.comparators import ComparatorManager


def pytest_configure(config):
    # Ensure set_path fixture runs before all tests.
    set_path()


@pytest.fixture(autouse=True, scope="session")
def environ():
    # Ensure normalize_environment fixture runs before each test.
    normalize_environment()


@pytest.fixture(autouse=True)
def reload_comparators():
    # Reload Comparators after every test so we are always in a consistent
    # state
    ComparatorManager().reload()


@pytest.fixture(autouse=True)
def reset_progress():
    ProgressManager().reset()



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2017-2020 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import struct

from diffoscope.difference import Difference

from .utils.file import File


class FontconfigCacheFile(File):
    DESCRIPTION = "FreeDesktop Fontconfig cache files"
    FILE_TYPE_HEADER_PREFIX = struct.pack("<H", 0xFC04)
    FILE_EXTENSION_SUFFIX = {"-le64.cache-4"}

    def compare_details(self, other, source=None):
        return [
            Difference.from_text(
                describe_cache_file(self.path),
                describe_cache_file(other.path),
                self.path,
                other.path,
            )
        ]


def describe_cache_file(filename):
    fmt = "<IIQQQQQQQ"
    fields = (
        "magic",
        "version",
        "size",
        "dir",
        "dirs",
        "dirs_count",
        "set",
        "checksum",
        "checksum_nano",
    )

    with open(filename, "rb") as f:
        data = struct.unpack(fmt, f.read(struct.calcsize(fmt)))
        kwargs = dict(zip(fields, data))

        kwargs["dir_name"] = read_null_terminated_string(f, kwargs["dir"])

    return """
struct FcCache {{
    unsigned int    magic = 0x{magic:08X};  /* FC_CACHE_MAGIC_MMAP or FC_CACHE_ALLOC */
    int             version = {version};  /* FC_CACHE_VERSION_NUMBER */
    intptr_t        size = {size};  /* size of file */
    intptr_t        dir = 0x{dir};  /* offset to dir name ("{dir_name}") */
    intptr_t        dirs = 0x{dirs:08X};  /* offset to subdirs */
    int             dirs_count = {dirs_count};  /* number of subdir strings */
    intptr_t        set = 0x{set:08X};  /* offset to font set */
    int             checksum = {checksum};  /* checksum of directory state */
    int64_t         checksum_nano = {checksum_nano};  /* checksum of directory state */
}};
""".format(
        **kwargs
    )


def read_null_terminated_string(fileobj, offset=None):
    result = ""

    if offset is not None:
        fileobj.seek(offset)

    while True:
        x = fileobj.read(1).decode("ascii")
        if x in ("", "\0"):
            break
        result += x

    return result



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2016-2022 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import os
import re
import abc
import magic
import logging
import subprocess

from diffoscope.exc import (
    RequiredToolNotFound,
    OutputParsingError,
    ContainerExtractionError,
)
from diffoscope.tools import tool_required
from diffoscope.utils import format_cmdline, format_class
from diffoscope.config import Config
from diffoscope.profiling import profile
from diffoscope.difference import Difference

try:
    import tlsh
except ImportError:  # noqa
    tlsh = None

SMALL_FILE_THRESHOLD = 65536  # 64 kiB

logger = logging.getLogger(__name__)


def path_apparent_size(path=".", visited=None):
    # should output the same as `du --apparent-size -bs "$path"`
    if not visited:
        stat = os.stat(path, follow_symlinks=False)
        visited = {stat.st_ino: stat.st_size}
    if os.path.isdir(path) and not os.path.islink(path):
        for entry in os.scandir(path):
            inode = entry.inode()
            if inode in visited:
                continue
            visited[inode] = entry.stat(follow_symlinks=False).st_size
            if entry.is_dir(follow_symlinks=False):
                path_apparent_size(entry.path, visited)
    return sum(visited.values())


def _run_tests(fold, tests):
    return fold(t(y, x) for x, t, y in tests)


class File(metaclass=abc.ABCMeta):
    """
    There are two Python modules named 'magic'. One of these ships with the
    file upstream under the python/ directory and PyPI's `python-magic` and
    they have slightly different interfaces despite attempts at a compatibility
    layer.
    """

    # Try and detect the instance of the libmagic shared library (loaded via
    # ctypes) used by the magic.py shipped with file.
    #
    if hasattr(magic, "_libraries"):

        @classmethod
        def guess_file_type(cls, path):
            if not hasattr(cls, "_mimedb"):
                cls._mimedb = magic.open(magic.NONE)
                cls._mimedb.load()
            return cls._mimedb.file(
                path.encode("utf-8", errors="surrogateescape")
            )

        @classmethod
        def guess_encoding(cls, path):
            if not hasattr(cls, "_mimedb_encoding"):
                cls._mimedb_encoding = magic.open(magic.MAGIC_MIME_ENCODING)
                cls._mimedb_encoding.load()
            return cls._mimedb_encoding.file(path)

    else:
        # Assume we are using python-magic

        @classmethod
        def guess_file_type(cls, path):
            if not hasattr(cls, "_mimedb"):
                cls._mimedb = magic.Magic()
            return maybe_decode(cls._mimedb.from_file(path))

        @classmethod
        def guess_encoding(cls, path):
            if not hasattr(cls, "_mimedb_encoding"):
                cls._mimedb_encoding = magic.Magic(mime_encoding=True)
            return maybe_decode(cls._mimedb_encoding.from_file(path))

    def __init__(self, container=None):
        self._comments = []
        self._container = container

    def __repr__(self):
        return f"<{self.__class__} {self.name}>"

    # This should return a path that allows to access the file content
    @property
    @abc.abstractmethod
    def path(self):
        raise NotImplementedError()

    # Remove any temporary data associated with the file. The function
    # should be idempotent and work during the destructor.
    def cleanup(self):
        if hasattr(self, "_as_container"):
            del self._as_container

    def __del__(self):
        self.cleanup()

    FILE_EXTENSION_SUFFIX = None
    FILE_TYPE_RE = None
    FILE_TYPE_HEADER_PREFIX = None  # max 16 bytes

    @classmethod
    def recognizes(cls, file):
        """Check if a file's type matches the one represented by this class.

        The default test returns True if the file matches these tests:

        (cls.FILE_TYPE_RE OR
         cls.FILE_TYPE_HEADER_PREFIX) AND
        (cls.FILE_EXTENSION_SUFFIX)

        If any test is None then the test is ignored and effectively deleted
        from the above definition.

        By default, the tests are all None and the test returns False for all
        files. Subclasses may override them with specific values, or override
        this method to implement a totally different test.
        """
        # The structure below allows us to construct a boolean tree of tests
        # that can be combined with all() and any(). Tests that are not defined
        # for a class are filtered out, so that we don't get into a "vacuous
        # truth" situation like a naive all([]) invocation would give.

        file_type_tests = [
            test
            for test in (
                (
                    cls.FILE_TYPE_RE,
                    lambda m, t: t.search(m),
                    file.magic_file_type,
                ),
                (
                    cls.FILE_TYPE_HEADER_PREFIX,
                    bytes.startswith,
                    file.file_header,
                ),
            )
            if test[0]
        ]  # filter out undefined tests

        all_tests = [
            test
            for test in (
                (cls.FILE_EXTENSION_SUFFIX, cls.any_endswith, file.name),
                (file_type_tests, _run_tests, any),
            )
            if test[0]
        ]  # filter out undefined tests, inc. file_type_tests if it's empty

        return _run_tests(all, all_tests) if all_tests else False

    FALLBACK_FILE_EXTENSION_SUFFIX = None
    FALLBACK_FILE_TYPE_HEADER_PREFIX = None  # max 16 bytes

    @classmethod
    def fallback_recognizes(cls, file):
        """This is checked if the file could not be identified by recognizes().
        This helps to work around bugs in file(1), see Debian bug #876316.

        The default test returns True if the file matches these tests:

        (cls.FALLBACK_FILE_EXTENSION_SUFFIX AND cls.FILE_EXTENSION_SUFFIX) AND
        (cls.FALLBACK_FILE_TYPE_HEADER_PREFIX AND cls.FILE_TYPE_HEADER_PREFIX)

        We also AND-compare with the non-fallback versions to ensure that
        subclasses don't "accidentally match" (e.g. IpkFile vs GzipFile).
        """
        if cls.recognizes.__func__ != File.recognizes.__func__:
            # If the class has overridden the default recognizes() then the
            # logic below about AND-comparing with the non-fallback versions is
            # not valid, they have to re-implement it
            return False

        all_tests = [
            test
            for test in (
                (
                    cls.FALLBACK_FILE_EXTENSION_SUFFIX,
                    cls.any_endswith,
                    file.name,
                ),
                (cls.FILE_EXTENSION_SUFFIX, cls.any_endswith, file.name),
                (
                    cls.FALLBACK_FILE_TYPE_HEADER_PREFIX,
                    bytes.startswith,
                    file.file_header,
                ),
                (
                    cls.FILE_TYPE_HEADER_PREFIX,
                    bytes.startswith,
                    file.file_header,
                ),
            )
            if test[0]
        ]  # filter out undefined tests, inc. file_type_tests if it's empty

        return _run_tests(all, all_tests) if all_tests else False

    @staticmethod
    def any_endswith(val, candidates):
        """
        Return true iff `val` ends with any string present in `candidates`
        iterable.
        """

        for x in candidates:
            if val.endswith(x):
                return True

        return False

    # This might be different from path and is used to do file extension matching
    @property
    def name(self):
        return self._name

    @property
    def container(self):
        return self._container

    CONTAINER_CLASSES = []

    @property
    def as_container(self):
        klasses = self.__class__.CONTAINER_CLASSES

        if not klasses:
            if hasattr(self, "_other_file"):
                return self._other_file.__class__.CONTAINER_CLASSES[0](self)
            return None

        if hasattr(self, "_as_container"):
            return self._as_container

        self._as_container = None

        # Try each container class in turn, returning the first one that
        # instantiates without error.
        for klass in klasses:
            formatted_class = format_class(
                klass, strip="diffoscope.comparators."
            )

            logger.debug(
                "Instantiating a %s for %s",
                formatted_class,
                self.name,
            )
            try:
                self._as_container = klass(self)

                return self._as_container
            except RequiredToolNotFound as exc:
                logger.debug(
                    "Cannot instantiate a %s; missing tool %s",
                    formatted_class,
                    exc.operation,
                )
                try:
                    infix = type(self).DESCRIPTION
                except AttributeError:
                    infix = "this file format"
                msg = (
                    "Format-specific differences are supported for {}.".format(
                        infix
                    )
                )
                self._comments.append(exc.get_comment(msg))

    @property
    def progress_name(self):
        x = self._name

        return x[1:] if x.startswith("./") else x

    @property
    def magic_file_type(self):
        if not hasattr(self, "_magic_file_type"):
            self._magic_file_type = File.guess_file_type(self.path)
        return self._magic_file_type

    @property
    def file_header(self):
        if not hasattr(self, "_file_header"):
            with open(self.path, "rb") as f:
                self._file_header = f.read(16)
        return self._file_header

    @property
    def file_type(self):
        for x, y in (
            (self.is_device, "device"),
            (self.is_symlink, "symlink"),
            (self.is_directory, "directory"),
        ):
            if x():
                return y

        return "file"

    if tlsh:

        @property
        def fuzzy_hash(self):
            def calc():
                # tlsh is not meaningful with files smaller than 512 bytes
                try:
                    if os.stat(self.path).st_size < 512:
                        return None
                except FileNotFoundError:
                    # eg. invalid symlink
                    return None

                h = tlsh.Tlsh()
                with open(self.path, "rb") as f:
                    for buf in iter(lambda: f.read(32768), b""):
                        h.update(buf)
                h.final()

                try:
                    return h.hexdigest()
                except ValueError:
                    # File must contain a certain amount of randomness.
                    return None

            if not hasattr(self, "_fuzzy_hash"):
                self._fuzzy_hash = calc()
            return self._fuzzy_hash

    @abc.abstractmethod
    def is_directory():
        raise NotImplementedError()

    @abc.abstractmethod
    def is_symlink():
        raise NotImplementedError()

    @abc.abstractmethod
    def is_device():
        raise NotImplementedError()

    def compare_bytes(self, other, source=None):
        from .compare import compare_binary_files

        # Don't attempt to compare directories with any other type as binaries
        if os.path.isdir(self.path) or os.path.isdir(other.path):
            return Difference.from_text(
                "type: {}".format(self.file_type),
                "type: {}".format(other.file_type),
                self.name,
                other.name,
                source,
            )

        return compare_binary_files(self, other, source)

    @staticmethod
    def _mangle_file_type(val):
        # Strip off trailing (eg.) "original size modulo 2^32 671" from
        # gzip compressed data as this is just a symptom of the contents itself
        # changing that will be reflected elsewhere.
        if val.startswith("gzip compressed data"):
            val = re.compile(r", original size modulo 2\^\d+ \d+$").sub(
                "", val
            )

        return val

    def _compare_using_details(self, other, source):
        details = []
        difference = Difference(self.name, other.name, source=source)

        if hasattr(self, "compare_details"):
            details.extend(self.compare_details(other, source))
        if self.as_container:
            if self.as_container.auto_diff_metadata:
                details.extend(
                    [
                        Difference.from_text(
                            self._mangle_file_type(self.magic_file_type),
                            self._mangle_file_type(other.magic_file_type),
                            self,
                            other,
                            source="filetype from file(1)",
                        ),
                        Difference.from_text(
                            self.__class__.__name__,
                            other.__class__.__name__,
                            self,
                            other,
                            source="filetype from diffoscope",
                        ),
                    ]
                )
            # Don't recurse forever on archive quines, etc.
            depth = self._as_container.depth
            no_recurse = depth >= Config().max_container_depth
            if no_recurse:
                msg = "Reached max container depth ({})".format(depth)
                logger.debug(msg)
                difference.add_comment(msg)
            elif Config().timeout_exceeded():
                msg = "Timeout exceeded; details may be incomplete."
                logger.debug(msg)
                difference.add_comment(msg)
                no_recurse = True

            details.extend(
                self.as_container.compare(
                    other.as_container, no_recurse=no_recurse
                )
            )

        details = [x for x in details if x]
        if not details:
            return None
        difference.add_details(details)

        return difference

    def has_same_content_as(self, other):
        logger.debug(
            "has_same_content(%s, %s)", self.path or "-", other.path or "-"
        )
        if os.path.isdir(self.path) or os.path.isdir(other.path):
            return False
        # try comparing small files directly first
        try:
            my_size = os.path.getsize(self.path)
            other_size = os.path.getsize(other.path)
        except OSError:
            # files not readable (e.g. broken symlinks) or something else,
            # just assume they are different
            return False
        if my_size != other_size:
            return False

        # Compare from python the first bytes, and only if they are
        # identical then call external command.
        assert my_size == other_size
        try:
            with profile("command", "cmp (internal)"):
                with open(self.path, "rb") as file1, open(
                    other.path, "rb"
                ) as file2:
                    content1 = file1.read(SMALL_FILE_THRESHOLD)
                    content2 = file2.read(SMALL_FILE_THRESHOLD)
        except OSError:
            # one or both files could not be opened for some reason,
            # assume they are different
            return False

        if content1 != content2:
            return False
        if my_size <= SMALL_FILE_THRESHOLD:
            return True

        # Big files, same size, same first bytes
        return self.cmp_external(other)

    @tool_required("cmp")
    def cmp_external(self, other):
        cmdline = ("cmp", "-s", self.path, other.path)
        logger.debug("Executing: %s", " ".join(cmdline))

        with profile("command", "cmp (external)"):
            return subprocess.call(cmdline, close_fds=True) == 0

    # To be specialized directly, or by implementing compare_details
    def compare(self, other, source=None):
        if hasattr(self, "compare_details") or self.as_container:
            try:
                difference = self._compare_using_details(other, source)
                # no differences detected inside? let's at least do a binary diff
                if difference is None:
                    difference = self.compare_bytes(other, source=source)
                    if difference is None:
                        return None
                    try:
                        infix = type(self).DESCRIPTION
                    except AttributeError:
                        infix = "this file format"
                    suffix = ""
                    if self.magic_file_type != "data":
                        suffix = " file(1) reports: {}".format(
                            self.magic_file_type
                        )
                    difference.add_comment(
                        "Format-specific differences are supported for {} but "
                        "no file-specific differences were detected; falling "
                        "back to a binary diff.{}".format(infix, suffix)
                    )
            except subprocess.CalledProcessError as e:
                difference = self.compare_bytes(other, source=source)
                if difference is None:
                    return None

                # Include either stderr (prefered) or stdout in the hexdump
                # difference
                suffix = None
                for prefix, val in (
                    ("Standard output", e.stdout),
                    ("Standard error", e.stderr),
                ):
                    if not val:
                        continue
                    suffix = " {}:\n{}".format(
                        prefix,
                        re.sub(
                            r"^",
                            "    ",
                            val.decode("utf-8").strip(),
                            flags=re.MULTILINE,
                        ),
                    )

                    # Truncate output
                    max_len = 250
                    if len(suffix) > max_len:
                        suffix = "{}  [...]".format(suffix[:max_len])

                difference.add_comment(
                    "Command `{}` failed with exit code {}.{}".format(
                        format_cmdline(e.cmd),
                        e.returncode,
                        suffix or " (No output)",
                    )
                )
            except RequiredToolNotFound as e:
                difference = self.compare_bytes(other, source=source)
                if difference is None:
                    return None
                difference.add_comment(
                    e.get_comment("Falling back to binary comparison.")
                )
            except OutputParsingError as e:
                difference = self.compare_bytes(other, source=source)
                if difference is None:
                    return None
                difference.add_comment(
                    f"Error parsing output of `{e.operation}` for {e.object_class}"
                )
            except ContainerExtractionError as e:
                difference = self.compare_bytes(other, source=source)
                if difference is None:
                    return None
                difference.add_comment(
                    "Error extracting '{}', falling back to "
                    'binary comparison: "{}"'.format(e.pathname, e.wrapped_exc)
                )

            # Append any miscellaneous comments for this file.
            for x in getattr(self, "_comments", []):
                difference.add_comment(x)

            return difference
        return self.compare_bytes(other, source)

    def add_comment(self, msg):
        self._comments.append(msg)


def maybe_decode(s):
    """
    Helper function to convert to bytes if necessary.
    """

    if type(s) is bytes:
        return s.decode("utf-8")
    return s



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2015      Jérémy Bobbio <lunar@debian.org>
# Copyright © 2016-2017 Mattia Rizzolo <mattia@debian.org>
# Copyright © 2019-2020 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import contextlib
import os
import re
import pytest

from diffoscope.comparators.binary import FilesystemFile
from diffoscope.comparators.directory import FilesystemDirectory
from diffoscope.comparators.utils.specialize import specialize

re_normalize_zeros = re.compile(
    r"^(?P<prefix>[ \-\+])(?P<offset>[0-9a-f]+)(?=: )", re.MULTILINE
)


def init_file(filename):
    # Ensure that any (specialized) FilesystemFile in tests have a
    # corresponding Container for any comparator that needs to know which
    # directory the file came from (eg. rdata)
    container = FilesystemDirectory(os.path.dirname(filename)).as_container

    return specialize(FilesystemFile(filename, container))


def init_fixture(filename):
    return pytest.fixture(lambda: init_file(filename))


def data(filename):
    return os.path.join(
        os.path.dirname(os.path.dirname(__file__)), "data", filename
    )


def get_data(filename):
    with open(data(filename), encoding="utf-8") as f:
        return f.read()


def assert_diff(difference, filename):
    # Assign seen and expected values to local variables to improve contextual
    # information in failed tests.
    seen = difference.unified_diff
    expected = get_data(filename)
    assert seen == expected


def assert_diff_startswith(difference, filename):
    haystack = difference.unified_diff
    needle = get_data(filename)
    assert haystack.startswith(needle)


# https://code.activestate.com/recipes/576620-changedirectory-context-manager/#c3
@contextlib.contextmanager
def cwd_data():
    """A context manager which changes the working directory to the given
    path, and then changes it back to its previous value on exit.
    """
    prev_cwd = os.getcwd()
    os.chdir(data(""))
    try:
        yield
    finally:
        os.chdir(prev_cwd)


def load_fixture(filename):
    return init_fixture(data(filename))


def normalize_zeros(s):
    # older xxd had one zero less.  Make sure there are always 8.
    def repl(x):
        return "{}{:08x}".format(x.group("prefix"), int(x.group("offset"), 16))

    return re_normalize_zeros.sub(repl, s)



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2016-2020 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import io
import os
import shlex
import logging
import binascii
import subprocess

from diffoscope.tools import tool_required
from diffoscope.exc import RequiredToolNotFound
from diffoscope.utils import exit_if_paths_do_not_exist
from diffoscope.config import Config
from diffoscope.excludes import any_excluded
from diffoscope.profiling import profile
from diffoscope.difference import Difference

from ..missing_file import MissingFile

from .command import Command
from .specialize import specialize


logger = logging.getLogger(__name__)


class Xxd(Command):
    @tool_required("xxd")
    def cmdline(self):
        return ["xxd", self.path]


def compare_root_paths(path1, path2):
    from ..directory import (
        FilesystemDirectory,
        FilesystemFile,
        compare_directories,
        compare_meta,
    )

    if not Config().new_file:
        exit_if_paths_do_not_exist(path1, path2)
    if any_excluded(path1, path2):
        return None

    if os.path.isdir(path1) and os.path.isdir(path2):
        return compare_directories(path1, path2)

    container1 = FilesystemDirectory(os.path.dirname(path1)).as_container
    file1 = specialize(FilesystemFile(path1, container=container1))
    container2 = FilesystemDirectory(os.path.dirname(path2)).as_container
    file2 = specialize(FilesystemFile(path2, container=container2))
    difference = compare_files(file1, file2)

    if Config().exclude_directory_metadata in ("no", "recursive"):
        meta = compare_meta(path1, path2)
        if meta:
            # Create an "empty" difference so we have something to attach file
            # metadata to.
            if difference is None:
                difference = Difference(file1.name, file2.name)
            difference.add_details(meta)

    # We cannot use isinstance() etc. because these instances are
    # abc-instantiated abstract base classes. We cannot even compare
    # DESCRIPTION attributes either, as the specialize() call above may not
    # have returned a specific instance.
    if difference is None and str(type(file1)) != str(type(file2)):
        difference = Difference(file1.name, file2.name)
        difference.add_comment(
            "Types of files differ; human-readable metadata may match literal file contents."
        )

    return difference


def compare_files(file1, file2, source=None, diff_content_only=False):
    logger.debug(
        "Comparing %s (%s) and %s (%s)",
        file1.name or "-",
        file1.__class__.__name__,
        file2.name or "-",
        file2.__class__.__name__,
    )

    if any_excluded(file1.name, file2.name):
        return None

    # Specialize the files first so "has_same_content_as" can be overridden
    # by subclasses
    with profile("specialize", "specialize"):
        specialize(file1)
        specialize(file2)

    force_details = Config().force_details
    with profile("has_same_content_as", file1):
        has_same_content = file1.has_same_content_as(file2)

    if has_same_content:
        if not force_details:
            logger.debug(
                "has_same_content_as returned True; skipping further comparisons"
            )
            return None
        if diff_content_only:
            return None
    elif diff_content_only:
        return Difference(file1.name, file2.name, comment="Files differ")

    call_difftool(file1, file2)

    if isinstance(file1, MissingFile):
        file1.other_file = file2
    elif isinstance(file2, MissingFile):
        file2.other_file = file1
    elif (
        (file1.__class__.__name__ != file2.__class__.__name__)
        and (file1.as_container is None or file2.as_container is None)
        and (not file1.is_directory() and not file2.is_directory())
    ):
        return file1.compare_bytes(file2, source)
    with profile("compare_files (cumulative)", file1):
        return file1.compare(file2, source)


def call_difftool(file1, file2):
    """
    Call an external difftool one-by-one, similar to git-difftool(1).
    """

    if Config().difftool is None:
        return

    a = "/dev/null" if isinstance(file1, MissingFile) else file1.path
    b = "/dev/null" if isinstance(file2, MissingFile) else file2.path

    if os.path.isdir(a) or os.path.isdir(b):
        return

    cmd = " ".join((Config().difftool, shlex.quote(a), shlex.quote(b)))
    logger.debug("Calling external command: %s", " ".join(cmd))
    subprocess.call(cmd, shell=True)


def compare_binary_files(file1, file2, source=None):
    try:
        if source is None:
            source = [file1.name, file2.name]
        return Difference.from_operation(
            Xxd,
            file1.path,
            file2.path,
            source=source,
            has_internal_linenos=True,
        )
    except RequiredToolNotFound:
        hexdump1 = hexdump_fallback(file1.path)
        hexdump2 = hexdump_fallback(file2.path)
        comment = (
            "xxd not available in path. Falling back to Python hexlify.\n"
        )
        return Difference.from_text(
            hexdump1, hexdump2, file1.name, file2.name, source, comment
        )


def hexdump_fallback(path):
    hexdump = io.StringIO()
    with open(path, "rb") as f:
        for buf in iter(lambda: f.read(32), b""):
            hexdump.write(
                "{}\n".format(binascii.hexlify(buf).decode("us-ascii"))
            )
    return hexdump.getvalue()



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2015 Jérémy Bobbio <lunar@debian.org>
# Copyright © 2015-2017, 2020 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import pytest

from diffoscope.comparators.cpio import CpioFile

from ..utils.data import load_fixture, get_data
from ..utils.tools import skip_unless_tools_exist
from ..utils.nonexisting import assert_non_existing

cpio1 = load_fixture("test1.cpio")
cpio2 = load_fixture("test2.cpio")


def test_identification(cpio1, cpio2):
    assert isinstance(cpio1, CpioFile)
    assert isinstance(cpio2, CpioFile)


def test_no_differences(cpio1):
    difference = cpio1.compare(cpio1)
    assert difference is None


@pytest.fixture
def differences(cpio1, cpio2):
    return cpio1.compare(cpio2).details


@skip_unless_tools_exist("cpio")
def test_listing(differences):
    expected_diff = get_data("cpio_listing_expected_diff")
    assert differences[0].unified_diff == expected_diff


@skip_unless_tools_exist("cpio")
def test_symlink(differences):
    assert differences[1].source1 == "dir/link"
    assert differences[1].comment == "symlink"
    expected_diff = get_data("symlink_expected_diff")
    assert differences[1].unified_diff == expected_diff


@skip_unless_tools_exist("cpio")
def test_compressed_files(differences):
    assert differences[2].source1 == "dir/text"
    assert differences[2].source2 == "dir/text"
    expected_diff = get_data("text_ascii_expected_diff")
    assert differences[2].unified_diff == expected_diff


@skip_unless_tools_exist("cpio")
def test_compare_non_existing(monkeypatch, cpio1):
    assert_non_existing(monkeypatch, cpio1)



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2016 Emanuel Bronshtein <e3amn2l@gmx.com>
# Copyright © 2017 Vagrant Cascadian <vagrant@debian.org>
# Copyright © 2017, 2019-2020 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import re
import pytest
import subprocess

from diffoscope.config import Config
from diffoscope.comparators.dtb import DeviceTreeFile
from diffoscope.comparators.missing_file import MissingFile

from ..utils.data import load_fixture, assert_diff
from ..utils.tools import skip_unless_tools_exist, skip_unless_tool_is_at_least

# Generated by: dtc --in-format=dts --out-format=dtb --out=devicetree1.dtb devicetree1.dts
dtb1 = load_fixture("devicetree1.dtb")
# Generated by: dtc --in-format=dts --out-format=dtb --out=devicetree2.dtb devicetree2.dts
dtb2 = load_fixture("devicetree2.dtb")


def fdtdump_version():
    out = subprocess.check_output(
        ("fdtdump", "--version"), stderr=subprocess.STDOUT
    ).decode("utf-8")

    m = re.search(r"Version: DTC (?P<version>\d+\.\d+\.\d+)", out)
    if m is None:
        raise ValueError("Error parsing `fdtdump --version` output")

    return m.group("version")


def test_identification(dtb1):
    assert isinstance(dtb1, DeviceTreeFile)


def test_no_differences(dtb1):
    difference = dtb1.compare(dtb1)
    assert difference is None


@pytest.fixture
def differences(dtb1, dtb2):
    return dtb1.compare(dtb2).details


@skip_unless_tool_is_at_least("fdtdump", fdtdump_version, "1.4.2")
def test_diff(differences):
    assert_diff(differences[0], "devicetree_expected_diff")


@skip_unless_tools_exist("fdtdump")
def test_compare_non_existing(monkeypatch, dtb1):
    monkeypatch.setattr(Config(), "new_file", True)
    difference = dtb1.compare(MissingFile("/nonexisting", dtb1))
    assert difference.source2 == "/nonexisting"
    assert len(difference.details) > 0



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2015 Daniel Kahn Gillmor <dkg@fifthhorseman.net>
# Copyright © 2015 Jérémy Bobbio <lunar@debian.org>
# Copyright © 2015 Paul Gevers <elbrus@debian.org>
# Copyright © 2016-2020 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import os
import re
import logging
import subprocess

from diffoscope.tools import tool_required
from diffoscope.profiling import profile
from diffoscope.difference import Difference

from .utils.file import File
from .utils.command import Command, our_check_output

logger = logging.getLogger(__name__)


class Ppudump(Command):
    @tool_required("ppudump")
    def cmdline(self):
        return ["ppudump", self.path]

    def env(self):
        # ppudump will return times using the local timezone which is not ideal
        # to investigate files. TZ environment variable can be used to enforce UTC.
        # Currently there is no fpc release yet that includes the TC environment
        # variable, but it looks for timezone definitions in the directory
        # specified by TZDIR. So let's set it to a non-existent directory
        # so we get UTC output even when the system timezone is set otherwise.
        env = dict(os.environ)
        env["TZ"] = ":UTC"
        env["TZDIR"] = "/nonexistent"
        return env

    def filter(self, line):
        if re.match(
            r"^Analyzing %s \(v[0-9]+\)$" % re.escape(self.path),
            line.decode("utf-8", errors="ignore"),
        ):
            return b""
        return line


class PpuFile(File):
    DESCRIPTION = "FreePascal files (.ppu)"
    FILE_EXTENSION_SUFFIX = {".ppu"}

    @classmethod
    def recognizes(cls, file):
        if not super().recognizes(file):
            return False

        if not file.file_header.startswith(b"PPU"):
            return False

        ppu_version = file.file_header[3:6].decode("ascii", errors="ignore")

        if not hasattr(PpuFile, "ppu_version"):
            try:
                with profile("command", "ppudump"):
                    our_check_output(
                        ["ppudump", "-vh", file.path],
                        stderr=subprocess.STDOUT,
                    )
                PpuFile.ppu_version = ppu_version
            except subprocess.CalledProcessError as e:
                error = e.output.decode("utf-8", errors="ignore")
                m = re.search("Expecting PPU version ([0-9]+)", error)
                try:
                    PpuFile.ppu_version = m.group(1)
                except AttributeError:
                    if m is None:
                        PpuFile.ppu_version = None
                        logger.debug("Unable to read PPU version")
                    else:
                        raise
            except OSError:
                PpuFile.ppu_version = None
                logger.debug("Unable to read PPU version")

        if PpuFile.ppu_version != ppu_version:
            logger.debug(
                "ppudump version (%s) does not match header of %s (%s)",
                PpuFile.ppu_version,
                file.name,
                ppu_version,
            )
            return False

        return True

    def compare_details(self, other, source=None):
        return [Difference.from_operation(Ppudump, self.path, other.path)]



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2019 Jelle van der Waa <jelle@archlinux.org>
# Copyright © 2020 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import subprocess

import pytest

from diffoscope.comparators.zst import ZstFile
from diffoscope.comparators.binary import FilesystemFile
from diffoscope.comparators.utils.specialize import specialize

from ..utils.tools import skip_unless_tools_exist


def zst_fixture(prefix):
    @pytest.fixture
    def zstd(tmpdir):
        input_ = str(tmpdir.join(prefix))
        output = str(tmpdir.join("{}.zst".format(prefix)))

        with open(input_, "w") as f:
            f.write(prefix)

        subprocess.check_call(("zstd", "--quiet", "--no-progress", input_))

        return specialize(FilesystemFile(output))

    return zstd


zst1 = zst_fixture("test1")
zst2 = zst_fixture("test2")


@skip_unless_tools_exist("zstd")
def test_identification(zst1):
    assert isinstance(zst1, ZstFile)


@skip_unless_tools_exist("zstd")
def test_no_differences(zst1):
    difference = zst1.compare(zst1)
    assert difference is None


@pytest.fixture
def differences(zst1, zst2):
    return zst1.compare(zst2).details


@skip_unless_tools_exist("zstd")
def test_content_source(differences):
    assert differences[0].source1 == "test1"
    assert differences[0].source2 == "test2"



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright (C) 2017 Juliana Rodrigues <juliana.orod@gmail.com>
# Copyright © 2017, 2019-2020 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import sys
import pytest

from diffoscope.comparators.xml import XMLFile

from ..utils.data import load_fixture, get_data


xml_a = load_fixture("test1.xml")
xml_b = load_fixture("test2.xml")
invalid_xml = load_fixture("test_invalid.xml")


def test_identification(xml_a):
    assert isinstance(xml_a, XMLFile)


def test_invalid(invalid_xml):
    assert not isinstance(invalid_xml, XMLFile)


def test_no_differences(xml_a):
    assert xml_a.compare(xml_a) is None


@pytest.fixture
def differences(xml_a, xml_b):
    return xml_a.compare(xml_b).details


@pytest.mark.skipif(
    sys.version_info < (3, 8), reason="requires Python 3.8 or higher"
)
def test_diff(differences):
    expected_diff = get_data("test_xml_expected_diff")
    assert differences[0].unified_diff == expected_diff



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2015 Jérémy Bobbio <lunar@debian.org>
# Copyright © 2016-2017, 2019-2020 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import struct
import pytest
import re
import subprocess

from diffoscope.comparators.cbfs import CbfsFile
from diffoscope.comparators.binary import FilesystemFile
from diffoscope.comparators.utils.specialize import specialize

from ..utils.data import data, get_data
from ..utils.tools import skip_unless_tools_exist
from ..utils.nonexisting import assert_non_existing


TEST_FILE1_PATH = data("text_ascii1")
TEST_FILE2_PATH = data("text_ascii2")


@pytest.fixture
def rom1(tmpdir):
    path = str(tmpdir.join("coreboot1"))

    subprocess.check_call(
        ("cbfstool", path, "create", "-m", "x86", "-s", "32768")
    )

    subprocess.check_call(
        (
            "cbfstool",
            path,
            "add",
            "-f",
            TEST_FILE1_PATH,
            "-n",
            "test file",
            "-t",
            "raw",
        ),
    )

    return specialize(FilesystemFile(path))


@pytest.fixture
def rom2(tmpdir):
    size = 32768
    path = str(tmpdir.join("coreboot2.rom"))

    subprocess.check_call(
        ("cbfstool", path, "create", "-m", "x86", "-s", "%s" % size),
    )

    subprocess.check_call(
        (
            "cbfstool",
            path,
            "add",
            "-f",
            TEST_FILE2_PATH,
            "-n",
            "test file",
            "-t",
            "raw",
        ),
        shell=False,
    )

    # Remove the last 4 bytes to exercise the full header search
    buf = bytearray(size)
    with open(path, "rb") as f:
        f.readinto(buf)

    with open(path, "wb") as f:
        size = struct.unpack_from("!I", buf, offset=len(buf) - 4 - 32 + 8)[0]
        struct.pack_into("!I", buf, len(buf) - 4 - 32 + 8, size - 4)
        f.write(buf[:-4])

    return specialize(FilesystemFile(path))


@skip_unless_tools_exist("cbfstool")
def test_identification_using_offset(rom1):
    assert isinstance(rom1, CbfsFile)


@skip_unless_tools_exist("cbfstool")
def test_identification_without_offset(rom2):
    assert isinstance(rom2, CbfsFile)


@skip_unless_tools_exist("cbfstool")
def test_no_differences(rom1):
    difference = rom1.compare(rom1)
    assert difference is None


@pytest.fixture
def differences(rom1, rom2):
    difference = rom1.compare(rom2)
    return difference.details


@skip_unless_tools_exist("cbfstool")
def test_listing(differences):
    # Compares outputs of: "cbfstool $tmpdir/coreboot*.rom print"
    #
    # As the output of this command keeps changing slightly (see
    # https://salsa.debian.org/reproducible-builds/diffoscope/merge_requests/38/
    # and the git log of this file), perform only these basic coherence check.

    assert differences[0].source1.startswith("cbfstool")
    assert re.search(r"\+test file\s.*\sraw\s", differences[0].unified_diff)


@skip_unless_tools_exist("cbfstool")
def test_content(differences):
    assert differences[1].source1 == "test file"
    assert differences[1].source2 == "test file"
    expected_diff = get_data("text_ascii_expected_diff")
    assert differences[1].unified_diff == expected_diff


@skip_unless_tools_exist("cbfstool")
def test_compare_non_existing(monkeypatch, rom1):
    assert_non_existing(monkeypatch, rom1)



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2020 Conrad Ratschan <ratschance@gmail.com>
# Copyright © 2020 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import pytest

from diffoscope.comparators.binary import FilesystemFile
from diffoscope.comparators.uimage import UimageFile
from diffoscope.comparators.utils.specialize import specialize

from ..utils.data import load_fixture, get_data, assert_diff
from ..utils.tools import skip_unless_tools_exist, file_version_is_lt
from ..utils.nonexisting import assert_non_existing

cpio1 = load_fixture("test1.cpio")
cpio2 = load_fixture("test2.cpio")

# Bytes from a U-Boot legacy image header generated via mkimage
uboot_header1 = bytes(
    b"\x27\x05\x19\x56\xf8\x7a\xd2\x00"
    b"\x5f\xc1\x58\x2c\x00\x00\x04\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x34\x71\x61\xa5\x05\x07\x03\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00"
)


# Bytes from a U-Boot legacy image header generated via mkimage
uboot_header2 = bytes(
    b"\x27\x05\x19\x56\xe8\x66\x86\xf7"
    b"\x5f\xc1\x58\x44\x00\x00\x04\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\xc6\x3c\x4a\x06\x05\x07\x03\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00"
)


def uimage_fixture(header, prefix):
    @pytest.fixture
    def uboot_cpio(tmp_path, cpio1, cpio2):
        if prefix == "test1":
            cpio = cpio1
        else:
            cpio = cpio2
        cpio_uboot = tmp_path / "{}.cpio.uboot".format(prefix)
        with open(cpio_uboot, "wb") as t_fp:
            t_fp.write(header)
            with open(cpio.path, "rb") as r_fp:
                t_fp.write(bytes(r_fp.read()))
        return specialize(FilesystemFile(str(cpio_uboot)))

    return uboot_cpio


uboot_cpio1 = uimage_fixture(uboot_header1, "test1")
uboot_cpio2 = uimage_fixture(uboot_header2, "test2")


def test_identification(uboot_cpio1, uboot_cpio2):
    assert isinstance(uboot_cpio1, UimageFile)
    assert isinstance(uboot_cpio2, UimageFile)


def test_no_differences(uboot_cpio1):
    difference = uboot_cpio1.compare(uboot_cpio1)
    assert difference is None


@pytest.fixture
def differences(uboot_cpio1, uboot_cpio2):
    return uboot_cpio1.compare(uboot_cpio2).details


@pytest.fixture
def nested_differences(uboot_cpio1, uboot_cpio2):
    return uboot_cpio1.compare(uboot_cpio2).details[1].details


def test_file_differences(differences):
    filename = "uimage_expected_diff"
    # file-5.41 slightly changed the output format by dropping leading 0x.
    if file_version_is_lt("5.41"):
        filename = "uimage_expected_diff_pre_5_41"

    assert_diff(differences[0], filename)


@skip_unless_tools_exist("cpio")
def test_nested_listing(nested_differences):
    expected_diff = get_data("cpio_listing_expected_diff")
    assert nested_differences[0].unified_diff == expected_diff


@skip_unless_tools_exist("cpio")
def test_nested_symlink(nested_differences):
    assert nested_differences[1].source1 == "dir/link"
    assert nested_differences[1].comment == "symlink"
    expected_diff = get_data("symlink_expected_diff")
    assert nested_differences[1].unified_diff == expected_diff


@skip_unless_tools_exist("cpio")
def test_nested_compressed_files(nested_differences):
    assert nested_differences[2].source1 == "dir/text"
    assert nested_differences[2].source2 == "dir/text"
    expected_diff = get_data("text_ascii_expected_diff")
    assert nested_differences[2].unified_diff == expected_diff


@skip_unless_tools_exist("cpio")
def test_compare_non_existing(monkeypatch, uboot_cpio1):
    assert_non_existing(monkeypatch, uboot_cpio1)



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2017-2018, 2020 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import os
import sys
import json
import pytest

from diffoscope.main import main
from diffoscope.progress import ProgressManager, StatusFD

from .utils.tools import skip_unless_module_exists
from .utils.versions import Version


TEST_TAR1_PATH = os.path.join(os.path.dirname(__file__), "data", "test1.tar")
TEST_TAR2_PATH = os.path.join(os.path.dirname(__file__), "data", "test2.tar")


def run(capsys, *args):
    with pytest.raises(SystemExit) as exc:
        main(args)

    out, err = capsys.readouterr()

    return exc.value.code, out, err


def progressbar_version():
    import progressbar

    return progressbar.__version__


@pytest.fixture
def progressbar_err():
    # This fixture returns an expected error message depending on progressbar
    # version, from the lowest to the highest.
    expected_err = {"3.19": "ETA", "3.34": ""}

    actual_ver = progressbar_version()

    for k, v in expected_err.items():
        if Version(actual_ver) < Version(k):
            return v

    return ""


@skip_unless_module_exists("progressbar")
def test_progress(capsys, progressbar_err):
    ret, _, err = run(capsys, TEST_TAR1_PATH, TEST_TAR2_PATH, "--progress")

    assert ret == 1
    assert progressbar_err in err


def test_status_fd(capsys):
    ProgressManager().register(StatusFD(sys.stderr))

    ret, _, err = run(capsys, TEST_TAR1_PATH, TEST_TAR2_PATH)

    assert ret == 1

    # Parse lines and ensure we emitted at least one line
    output = [json.loads(x) for x in err.splitlines()]
    assert output

    # Ensure each line is valid
    for x in output:
        assert "msg" in x
        assert x["current"] <= x["total"]

    # Last line should mark us as "complete"
    assert output[-1]["current"] == output[-1]["total"]



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2016-2021 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import abc
import uuid
import os.path
import logging
import itertools
from collections import OrderedDict

from diffoscope.config import Config
from diffoscope.difference import Difference
from diffoscope.excludes import filter_excludes
from diffoscope.progress import Progress

from ..missing_file import MissingFile, AbstractMissingType

from .file import path_apparent_size
from .fuzzy import perform_fuzzy_matching

NO_COMMENT = None

logger = logging.getLogger(__name__)


class Container(metaclass=abc.ABCMeta):
    auto_diff_metadata = True

    def __new__(cls, source):
        if isinstance(source, MissingFile):
            new = super(Container, MissingContainer).__new__(MissingContainer)
            new.__init__(source)
            return new

        return super(Container, cls).__new__(cls)

    def __init__(self, source):
        self._source = source

        # Keep a count of how "nested" we are
        self.depth = 0
        if hasattr(source, "container") and source.container is not None:
            self.depth = source.container.depth + 1

    @property
    def source(self):
        return self._source

    @abc.abstractmethod
    def get_member_names(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def get_member(self, member_name):
        raise NotImplementedError()

    def get_path_name(self, dest_dir):
        return os.path.join(dest_dir, str(uuid.uuid4()))

    def get_filtered_members(self):
        # If your get_member implementation is O(n) then this will be O(n^2)
        # cost. In such cases it is recommended to override this as well
        for name in filter_excludes(self.get_member_names()):
            yield name, self.get_member(name)

    def perform_fuzzy_matching(self, my_members, other_members):
        return perform_fuzzy_matching(my_members, other_members)

    def get_adjusted_members(self):
        """
        Returns an iterable of pairs. The key is what is used to match when
        comparing containers. This may be used to e.g. strip off version
        numbers, hashes, etc, efficiently for known file formats, so that we
        don't need to use the expensive tlsh "fuzzy-hashing" logic.

        Note that containers with 1 element are already force-compared against
        other containers with 1 element, so you don't need to override this
        method for those cases.
        """
        return self.get_filtered_members()

    def lookup_file(self, *names):
        """
        Try to fetch a specific file by digging in containers.
        """

        from .specialize import specialize

        name, remainings = names[0], names[1:]
        try:
            file = self.get_member(name)
        except KeyError:
            return None

        logger.debug("lookup_file(%s) -> %s", names, file)
        specialize(file)
        if not remainings:
            return file

        container = file.as_container
        if not container:
            return None

        return container.lookup_file(*remainings)

    def get_adjusted_members_sizes(self):
        for name, member in self.get_adjusted_members():
            if member.is_directory():
                size = 4096  # default "size" of a directory
            else:
                size = path_apparent_size(member.path)
            yield name, (member, size)

    def comparisons(self, other):
        my_members = OrderedDict(self.get_adjusted_members_sizes())
        other_members = OrderedDict(other.get_adjusted_members_sizes())
        total_size = sum(
            x[1]
            for x in itertools.chain(
                my_members.values(), other_members.values()
            )
        )
        # TODO: progress could be a bit more accurate here, give more weight to fuzzy-hashed files

        with Progress(total_size) as p:

            def prep_yield(my_name, other_name, comment=NO_COMMENT):
                my_member, my_size = my_members.pop(my_name)
                other_member, other_size = other_members.pop(other_name)
                p.begin_step(my_size + other_size, msg=my_member.progress_name)
                return my_member, other_member, comment

            # if both containers contain 1 element, compare these
            if len(my_members) == 1 and len(other_members) == 1:
                yield prep_yield(
                    next(iter(my_members.keys())),
                    next(iter(other_members.keys())),
                )
                return

            other_names = set(other_members.keys())
            # keep it sorted like my_members
            both_names = [
                name for name in my_members.keys() if name in other_names
            ]
            for name in both_names:
                yield prep_yield(name, name)

            for my_name, other_name, score in self.perform_fuzzy_matching(
                my_members, other_members
            ):
                score_display = score / 400.0 * 100
                comment = f"Files {score_display:.0f}% similar despite different names"
                if score == 0:
                    comment = "Files identical despite different names"
                yield prep_yield(my_name, other_name, comment)

            if Config().new_file:
                for my_member, my_size in my_members.values():
                    p.begin_step(my_size, msg=my_member.progress_name)
                    yield my_member, MissingFile(
                        "/dev/null", my_member
                    ), NO_COMMENT

                for other_member, other_size in other_members.values():
                    p.begin_step(other_size, msg=other_member.progress_name)
                    yield MissingFile(
                        "/dev/null", other_member
                    ), other_member, NO_COMMENT

    def compare(self, other, source=None, no_recurse=False):
        from .compare import compare_files
        from ..directory import compare_meta

        def compare_pair(file1, file2, comment):
            if Config().timeout_exceeded():
                difference = Difference(file1.name, file2.name)
                msg = "Timeout exceeded; details may be incomplete."
                difference.add_comment(msg)
                return difference

            difference = compare_files(
                file1, file2, source=None, diff_content_only=no_recurse
            )

            if isinstance(file1, AbstractMissingType) or isinstance(
                file2, AbstractMissingType
            ):
                # There is no need to compare metadata with a missing file,
                # as it doesn't make much sense
                meta_differences = []
            else:
                meta_differences = compare_meta(file1.name, file2.name)

            if meta_differences and not difference:
                difference = Difference(file1.path, file2.path)
            if difference:
                difference.add_details(meta_differences)

            if comment:
                if difference is None:
                    difference = Difference(file1.name, file2.name)
                difference.add_comment(comment)
            return difference

        return filter(
            None, itertools.starmap(compare_pair, self.comparisons(other))
        )


class MissingContainer(Container, AbstractMissingType):
    def get_member_names(self):
        return self.source.other_file.as_container.get_member_names()

    def get_member(self, member_name):
        return MissingFile("/dev/null")



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2016 Emanuel Bronshtein <e3amn2l@gmx.com>
# Copyright © 2016-2020 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import pytest
import subprocess
import re

from diffoscope.config import Config
from diffoscope.comparators.openssh import PublicKeyFile
from diffoscope.comparators.missing_file import MissingFile

from ..utils.data import load_fixture, get_data
from ..utils.tools import skip_unless_tools_exist, skip_unless_tool_is_at_least


# Generated by: ssh-keygen -t dsa -C "Test1"
opensshpubkey1 = load_fixture("test_openssh_pub_key1.pub")
# Generated by: ssh-keygen -t rsa -b 4096 -C "Test2"
opensshpubkey2 = load_fixture("test_openssh_pub_key2.pub")


def openssh_version():
    out = subprocess.check_output(("ssh", "-V"), stderr=subprocess.STDOUT)
    return re.match(r"OpenSSH_([a-zA-Z0-9._+-]+)", out.decode()).group(1)


def test_identification(opensshpubkey1):
    assert isinstance(opensshpubkey1, PublicKeyFile)


def test_no_differences(opensshpubkey1):
    difference = opensshpubkey1.compare(opensshpubkey1)
    assert difference is None


@pytest.fixture
def differences(opensshpubkey1, opensshpubkey2):
    return opensshpubkey1.compare(opensshpubkey2).details


@skip_unless_tool_is_at_least("ssh-keygen", openssh_version, "6.9")
def test_diff(differences):
    expected_diff = get_data("openssh_pub_key_expected_diff")
    assert differences[0].unified_diff == expected_diff


@skip_unless_tools_exist("ssh-keygen")
def test_compare_non_existing(monkeypatch, opensshpubkey1):
    monkeypatch.setattr(Config(), "new_file", True)
    difference = opensshpubkey1.compare(
        MissingFile("/nonexisting", opensshpubkey1)
    )
    assert difference.source2 == "/nonexisting"
    assert len(difference.details) > 0



# -*- coding: utf-8 -*-
#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2020 Conrad Ratschan <ratschance@gmail.com>
# Copyright © 2021 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.
import os
import subprocess

import pytest

from diffoscope.comparators.binary import FilesystemFile
from diffoscope.comparators.fit import FlattenedImageTreeFile
from diffoscope.comparators.utils.specialize import specialize

from ..utils.data import data, assert_diff, load_fixture
from ..utils.tools import skip_unless_tools_exist, skip_unless_tool_is_at_least
from ..utils.nonexisting import assert_non_existing

cpio1 = load_fixture("test1.cpio")
cpio2 = load_fixture("test2.cpio")


def dumpimage_version():
    return (
        subprocess.check_output(("dumpimage", "-V"))
        .decode("utf-8")
        .strip()
        .split(" ")[-1]
    )


def fit_fixture(prefix, entrypoint):
    @pytest.fixture
    def uboot_fit(tmp_path):
        cpio = data("{}.cpio".format(prefix))
        fit_out = tmp_path / "{}.itb".format(prefix)

        # Time must be controlled for reproducible FIT images
        time_env = os.environ.copy()
        time_env["SOURCE_DATE_EPOCH"] = "1609459200"
        time_env["TZ"] = "UTC"
        _ = subprocess.run(
            [
                "mkimage",
                "-f",
                "auto",
                "-A",
                "arm",
                "-O",
                "linux",
                "-T",
                "ramdisk",
                "-C",
                "none",
                "-e",
                entrypoint,
                "-d",
                cpio,
                fit_out,
            ],
            capture_output=True,
            env=time_env,
        )
        return specialize(FilesystemFile(str(fit_out)))

    return uboot_fit


uboot_fit1 = fit_fixture("test1", "0x1000")
uboot_fit2 = fit_fixture("test2", "0x2000")


@skip_unless_tools_exist("dumpimage")
@skip_unless_tools_exist("fdtdump")
def test_identification(uboot_fit1, uboot_fit2):
    assert isinstance(uboot_fit1, FlattenedImageTreeFile)
    assert isinstance(uboot_fit2, FlattenedImageTreeFile)


@skip_unless_tools_exist("dumpimage")
@skip_unless_tools_exist("fdtdump")
def test_no_differences(uboot_fit1):
    difference = uboot_fit1.compare(uboot_fit1)
    assert difference is None


@pytest.fixture
def differences(uboot_fit1, uboot_fit2):
    return uboot_fit1.compare(uboot_fit2).details


@pytest.fixture
def nested_differences(uboot_fit1, uboot_fit2):
    return uboot_fit1.compare(uboot_fit2).details[1].details


@skip_unless_tool_is_at_least("dumpimage", dumpimage_version, "2021.01")
@skip_unless_tools_exist("fdtdump")
def test_file_differences(differences):
    assert_diff(differences[0], "fit_expected_diff")


@skip_unless_tools_exist("cpio")
@skip_unless_tool_is_at_least("dumpimage", dumpimage_version, "2021.01")
@skip_unless_tools_exist("fdtdump")
def test_nested_listing(nested_differences):
    assert_diff(nested_differences[0], "cpio_listing_expected_diff")


@skip_unless_tools_exist("cpio")
@skip_unless_tool_is_at_least("dumpimage", dumpimage_version, "2021.01")
@skip_unless_tools_exist("fdtdump")
def test_nested_symlink(nested_differences):
    assert nested_differences[1].source1 == "dir/link"
    assert nested_differences[1].comment == "symlink"
    assert_diff(nested_differences[1], "symlink_expected_diff")


@skip_unless_tools_exist("cpio")
@skip_unless_tool_is_at_least("dumpimage", dumpimage_version, "2021.01")
@skip_unless_tools_exist("fdtdump")
def test_nested_compressed_files(nested_differences):
    assert nested_differences[2].source1 == "dir/text"
    assert nested_differences[2].source2 == "dir/text"
    assert_diff(nested_differences[2], "text_ascii_expected_diff")


@skip_unless_tools_exist("cpio")
@skip_unless_tool_is_at_least("dumpimage", dumpimage_version, "2021.01")
@skip_unless_tools_exist("fdtdump")
def test_compare_non_existing(monkeypatch, uboot_fit1):
    assert_non_existing(monkeypatch, uboot_fit1)



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2015 Jérémy Bobbio <lunar@debian.org>
# Copyright © 2015-2020 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import pytest

from diffoscope.config import Config
from diffoscope.comparators.ipk import IpkFile
from diffoscope.comparators.missing_file import MissingFile

from ..utils.data import load_fixture, get_data
from ..utils.tools import skip_unless_file_version_is_at_least


ipk1 = load_fixture("base-files_157-r45695_ar71xx.ipk")
ipk2 = load_fixture("base-files_157-r45918_ar71xx.ipk")


def test_identification(ipk1):
    assert isinstance(ipk1, IpkFile)


def test_no_differences(ipk1):
    difference = ipk1.compare(ipk1)
    assert difference is None


@pytest.fixture
def differences(ipk1, ipk2):
    return ipk1.compare(ipk2).details


@skip_unless_file_version_is_at_least("5.37")
def test_metadata(differences):
    assert differences[0].source1.startswith("filetype")
    expected_diff = get_data("ipk_metadata_expected_diff")
    assert differences[0].unified_diff == expected_diff


def test_compressed_files(differences):
    assert differences[1].details[1].source1 == "./data.tar.gz"
    assert differences[1].details[2].source1 == "./control.tar.gz"


def test_compare_non_existing(monkeypatch, ipk1):
    monkeypatch.setattr(Config(), "new_file", True)
    difference = ipk1.compare(MissingFile("/nonexisting", ipk1))
    assert difference.source2 == "/nonexisting"
    assert difference.details[-1].source2 == "/dev/null"



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2016-2017, 2020 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import pytest

from diffoscope.comparators.json import JSONFile

from ..utils.data import load_fixture, get_data
from ..utils.nonexisting import assert_non_existing


json1 = load_fixture("test1.json")
json2 = load_fixture("test2.json")
json3a = load_fixture("order1a.json")
json3b = load_fixture("order1b.json")
invalid_json = load_fixture("test_invalid.json")


def test_identification(json1):
    assert isinstance(json1, JSONFile)


def test_invalid(invalid_json):
    assert not isinstance(invalid_json, JSONFile)


def test_no_differences(json1):
    assert json1.compare(json1) is None


@pytest.fixture
def differences(json1, json2):
    return json1.compare(json2).details


def test_diff(differences):
    expected_diff = get_data("json_expected_diff")

    assert differences[0].unified_diff == expected_diff


def test_compare_non_existing(monkeypatch, json1):
    assert_non_existing(monkeypatch, json1)


def test_ordering_differences(json3a, json3b):
    diff = json3a.compare(json3b)
    assert diff.details[0]._comments == ["ordering differences only"]
    assert diff.details[0].unified_diff == get_data("order1.diff")



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2015 Jérémy Bobbio <lunar@debian.org>
# Copyright © 2016-2020 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import shutil
import pytest

from diffoscope.config import Config
from diffoscope.comparators import ComparatorManager
from diffoscope.comparators.binary import FilesystemFile
from diffoscope.comparators.missing_file import MissingFile
from diffoscope.comparators.utils.specialize import specialize

from ..utils.data import data, get_data
from ..utils.tools import skip_unless_module_exists

from ..utils.nonexisting import assert_non_existing

try:
    from diffoscope.comparators.debian import (
        DotChangesFile,
        DotDscFile,
        DotBuildinfoFile,
    )
except ImportError:
    from diffoscope.comparators.debian_fallback import (
        DotChangesFile,
        DotDscFile,
        DotBuildinfoFile,
    )

TEST_DOT_CHANGES_FILE1_PATH = data("test1.changes")
TEST_DOT_CHANGES_FILE2_PATH = data("test2.changes")
TEST_DOT_CHANGES_FILE3_PATH = data("test3.changes")
TEST_DOT_CHANGES_FILE4_PATH = data("test4.changes")
TEST_DOT_BUILDINFO_FILE1_PATH = data("test1.buildinfo")
TEST_DOT_BUILDINFO_FILE2_PATH = data("test2.buildinfo")
TEST_DEB_FILE1_PATH = data("test1.deb")
TEST_DEB_FILE2_PATH = data("test2.deb")


@pytest.fixture
def dot_changes1(tmpdir):
    tmpdir.mkdir("a")
    dot_changes_path = str(tmpdir.join("a/test_1.changes"))
    shutil.copy(TEST_DOT_CHANGES_FILE1_PATH, dot_changes_path)
    shutil.copy(TEST_DEB_FILE1_PATH, str(tmpdir.join("a/test_1_all.deb")))
    shutil.copy(
        TEST_DOT_BUILDINFO_FILE1_PATH, str(tmpdir.join("a/test_1.buildinfo"))
    )
    return specialize(FilesystemFile(dot_changes_path))


@pytest.fixture
def dot_changes2(tmpdir):
    tmpdir.mkdir("b")
    dot_changes_path = str(tmpdir.join("b/test_1.changes"))
    shutil.copy(TEST_DOT_CHANGES_FILE2_PATH, dot_changes_path)
    shutil.copy(TEST_DEB_FILE2_PATH, str(tmpdir.join("b/test_1_all.deb")))
    shutil.copy(
        TEST_DOT_BUILDINFO_FILE2_PATH, str(tmpdir.join("b/test_2.buildinfo"))
    )
    return specialize(FilesystemFile(dot_changes_path))


@pytest.fixture
def dot_changes3(tmpdir):
    tmpdir.mkdir("c")
    dot_changes_path = str(tmpdir.join("c/test_3.changes"))
    shutil.copy(TEST_DOT_CHANGES_FILE3_PATH, dot_changes_path)
    shutil.copy(TEST_DEB_FILE1_PATH, str(tmpdir.join("c/test_1_all.deb")))
    shutil.copy(
        TEST_DOT_BUILDINFO_FILE2_PATH, str(tmpdir.join("c/test_2.buildinfo"))
    )
    return specialize(FilesystemFile(dot_changes_path))


@pytest.fixture
def dot_changes4(tmpdir):
    tmpdir.mkdir("d")
    dot_changes_path = str(tmpdir.join("d/test_4.changes"))
    shutil.copy(TEST_DOT_CHANGES_FILE4_PATH, dot_changes_path)
    shutil.copy(TEST_DEB_FILE2_PATH, str(tmpdir.join("d/test_1_all.deb")))
    shutil.copy(
        TEST_DOT_BUILDINFO_FILE1_PATH, str(tmpdir.join("d/test_2.buildinfo"))
    )
    return specialize(FilesystemFile(dot_changes_path))


def test_dot_changes_identification(dot_changes1):
    assert isinstance(dot_changes1, DotChangesFile)


@skip_unless_module_exists("debian.deb822")
def test_dot_changes_invalid(tmpdir):
    tmpdir.mkdir("a")
    dot_changes_path = str(tmpdir.join("a/test_1.changes"))
    shutil.copy(TEST_DOT_CHANGES_FILE1_PATH, dot_changes_path)
    # we don't copy the referenced .deb
    identified = specialize(FilesystemFile(dot_changes_path))
    # ... but it is identified regardless
    assert isinstance(identified, DotChangesFile)


def test_dot_changes_no_differences(dot_changes1):
    difference = dot_changes1.compare(dot_changes1)
    assert difference is None


@pytest.fixture
def dot_changes_differences(dot_changes1, dot_changes2):
    difference = dot_changes1.compare(dot_changes2)
    assert difference.source2 == "/nonexisting"
    assert difference.details[-1].source2 == "/dev/null"


@pytest.fixture
def dot_changes_differences_identical_contents_and_identical_files(
    dot_changes1, dot_changes3
):
    difference = dot_changes1.compare(dot_changes3)
    return difference.details


@pytest.fixture
def dot_changes_differences_identical_contents_and_different_files(
    dot_changes1, dot_changes4
):
    difference = dot_changes1.compare(dot_changes4)
    return difference.details


@pytest.fixture
def dot_changes_differences_different_contents_and_identical_files(
    dot_changes2, dot_changes4
):
    difference = dot_changes4.compare(dot_changes2)
    return difference.details


@skip_unless_module_exists("debian.deb822")
def test_dot_changes_no_differences_exclude_buildinfo(
    dot_changes1, dot_changes3
):
    difference = dot_changes1.compare(dot_changes3)
    assert difference is None


@skip_unless_module_exists("debian.deb822")
def test_dot_changes_identical_contents_and_different_files(
    dot_changes_differences_identical_contents_and_different_files,
):
    assert dot_changes_differences_identical_contents_and_different_files[0]
    expected_diff = get_data(
        "dot_changes_identical_contents_and_different_files_expected_diff"
    )
    assert (
        dot_changes_differences_identical_contents_and_different_files[
            0
        ].unified_diff
        == expected_diff
    )


@skip_unless_module_exists("debian.deb822")
def test_dot_changes_different_contents_and_identical_files(
    dot_changes_differences_different_contents_and_identical_files,
):
    assert dot_changes_differences_different_contents_and_identical_files[0]
    assert dot_changes_differences_different_contents_and_identical_files[1]
    expected_diff_contents = get_data("dot_changes_description_expected_diff")
    expected_diff_files = get_data(
        "dot_changes_different_contents_and_identical_files_expected_diff"
    )
    assert (
        dot_changes_differences_different_contents_and_identical_files[
            0
        ].unified_diff
        == expected_diff_contents
    )
    assert (
        dot_changes_differences_different_contents_and_identical_files[
            1
        ].unified_diff
        == expected_diff_files
    )


TEST_DOT_DSC_FILE1_PATH = data("test1.dsc")
TEST_DOT_DSC_FILE2_PATH = data("test2.dsc")
TEST_DEB_SRC1_PATH = data("test1.debsrc.tar.gz")
TEST_DEB_SRC2_PATH = data("test2.debsrc.tar.gz")


@pytest.fixture
def dot_dsc1(tmpdir):
    tmpdir.mkdir("a")
    dot_dsc_path = str(tmpdir.join("a/test_1.dsc"))
    shutil.copy(TEST_DOT_DSC_FILE1_PATH, dot_dsc_path)
    shutil.copy(TEST_DEB_SRC1_PATH, str(tmpdir.join("a/test_1.tar.gz")))
    return specialize(FilesystemFile(dot_dsc_path))


@pytest.fixture
def dot_dsc2(tmpdir):
    tmpdir.mkdir("b")
    dot_dsc_path = str(tmpdir.join("b/test_1.dsc"))
    shutil.copy(TEST_DOT_DSC_FILE2_PATH, dot_dsc_path)
    shutil.copy(TEST_DEB_SRC2_PATH, str(tmpdir.join("b/test_1.tar.gz")))
    return specialize(FilesystemFile(dot_dsc_path))


def test_dot_dsc_identification(dot_dsc1):
    assert isinstance(dot_dsc1, DotDscFile)


@skip_unless_module_exists("debian.deb822")
def test_dot_dsc_no_associated_tar_gz(tmpdir, dot_dsc2):
    tmpdir.mkdir("a")
    dot_dsc_path = str(tmpdir.join("a/test_1.dsc"))
    shutil.copy(TEST_DOT_CHANGES_FILE1_PATH, dot_dsc_path)
    # we don't copy the referenced .tar.gz
    identified = specialize(FilesystemFile(dot_dsc_path))
    assert isinstance(identified, DotDscFile)


def test_dot_dsc_no_differences(dot_dsc1):
    difference = dot_dsc1.compare(dot_dsc1)
    assert difference is None


@pytest.fixture
def dot_dsc_differences(dot_dsc1, dot_dsc2):
    difference = dot_dsc1.compare(dot_dsc2)
    return difference.details


@skip_unless_module_exists("debian.deb822")
def test_dot_dsc_internal_diff(dot_dsc_differences):
    assert dot_dsc_differences[1].source1 == "test_1.tar.gz"


@skip_unless_module_exists("debian.deb822")
def test_dot_dsc_compare_non_existing(monkeypatch, dot_dsc1):
    monkeypatch.setattr(Config(), "new_file", True)
    difference = dot_dsc1.compare(MissingFile("/nonexisting", dot_dsc1))
    assert difference.source2 == "/nonexisting"
    assert difference.details[-1].source2 == "/dev/null"


@pytest.fixture
def dot_buildinfo1(tmpdir):
    tmpdir.mkdir("a")
    dot_buildinfo_path = str(tmpdir.join("a/test_1.buildinfo"))
    shutil.copy(TEST_DOT_BUILDINFO_FILE1_PATH, dot_buildinfo_path)
    shutil.copy(TEST_DOT_DSC_FILE1_PATH, str(tmpdir.join("a/test_1.dsc")))
    shutil.copy(TEST_DEB_FILE1_PATH, str(tmpdir.join("a/test_1_all.deb")))
    return specialize(FilesystemFile(dot_buildinfo_path))


@pytest.fixture
def dot_buildinfo2(tmpdir):
    tmpdir.mkdir("b")
    dot_buildinfo_path = str(tmpdir.join("b/test_1.buildinfo"))
    shutil.copy(TEST_DOT_BUILDINFO_FILE2_PATH, dot_buildinfo_path)
    shutil.copy(TEST_DOT_DSC_FILE2_PATH, str(tmpdir.join("b/test_1.dsc")))
    shutil.copy(TEST_DEB_FILE2_PATH, str(tmpdir.join("b/test_1_all.deb")))
    return specialize(FilesystemFile(dot_buildinfo_path))


def test_dot_buildinfo_identification(dot_buildinfo1):
    assert isinstance(dot_buildinfo1, DotBuildinfoFile)


@skip_unless_module_exists("debian.deb822")
def test_dot_buildinfo_no_deb(tmpdir):
    tmpdir.mkdir("a")
    dot_buildinfo_path = str(tmpdir.join("a/test_1.buildinfo"))
    shutil.copy(TEST_DOT_BUILDINFO_FILE1_PATH, dot_buildinfo_path)
    # we don't copy the referenced .deb
    identified = specialize(FilesystemFile(dot_buildinfo_path))
    assert isinstance(identified, DotBuildinfoFile)


def test_dot_buildinfo_no_differences(dot_buildinfo1):
    difference = dot_buildinfo1.compare(dot_buildinfo1)
    assert difference is None


@pytest.fixture
def dot_buildinfo_differences(dot_buildinfo1, dot_buildinfo2):
    difference = dot_buildinfo1.compare(dot_buildinfo2)
    return difference.details


@skip_unless_module_exists("debian.deb822")
def test_dot_buildinfo_internal_diff(dot_buildinfo_differences):
    assert dot_buildinfo_differences[1].source1 == "test_1_all.deb"


@skip_unless_module_exists("debian.deb822")
def test_dot_buildinfo_compare_non_existing(monkeypatch, dot_buildinfo1):
    assert_non_existing(monkeypatch, dot_buildinfo1)


def test_fallback_comparisons(monkeypatch):
    manager = ComparatorManager()
    monkeypatch.setattr(
        manager,
        "COMPARATORS",
        (
            ("debian_fallback.DotChangesFile",),
            ("debian_fallback.DotDscFile",),
            ("debian_fallback.DotBuildinfoFile",),
        ),
    )
    manager.reload()

    for a, b, expected_diff in (
        (
            "test1.changes",
            "test2.changes",
            "dot_changes_fallback_expected_diff",
        ),
        ("test1.dsc", "test2.dsc", "dot_dsc_fallback_expected_diff"),
        (
            "test1.buildinfo",
            "test2.buildinfo",
            "dot_buildinfo_fallback_expected_diff",
        ),
    ):
        # Re-specialize after reloading our Comparators
        file1 = specialize(FilesystemFile(data(a)))
        file2 = specialize(FilesystemFile(data(b)))

        assert file1.compare(file1) is None
        assert file2.compare(file2) is None
        assert file1.compare(file2).unified_diff == get_data(expected_diff)



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2015 Jérémy Bobbio <lunar@debian.org>
# Copyright © 2015-2020 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import os.path
import pytest
import subprocess

from os import mkdir, symlink
from tempfile import TemporaryDirectory

from diffoscope.tools import tool_required
from diffoscope.exc import RequiredToolNotFound
from diffoscope.difference import Difference
from diffoscope.comparators.binary import FilesystemFile
from diffoscope.comparators.utils.file import File
from diffoscope.comparators.missing_file import MissingFile
from diffoscope.comparators.utils.compare import Xxd

from ..utils.data import data, init_fixture, get_data, normalize_zeros
from ..utils.tools import (
    skip_unless_tools_exist,
    skip_unless_module_exists,
    file_version_is_lt,
)


TEST_FILE1_PATH = data("binary1")
TEST_FILE2_PATH = data("binary2")
TEST_ASCII_PATH = data("text_ascii1")
TEST_UNICODE_PATH = data("text_unicode1")
TEST_ISO8859_PATH = data("text_iso8859")


binary1 = init_fixture(TEST_FILE1_PATH)
binary2 = init_fixture(TEST_FILE2_PATH)


def test_same_content(binary1):
    assert binary1.has_same_content_as(binary1) is True


def test_not_same_content(binary1, binary2):
    assert binary1.has_same_content_as(binary2) is False


def test_guess_file_type():
    if file_version_is_lt("5.40"):
        assert File.guess_file_type(TEST_FILE1_PATH) == "data"
    else:
        assert File.guess_file_type(TEST_FILE1_PATH) == "OpenPGP Public Key"


def test_guess_encoding_binary():
    assert File.guess_encoding(TEST_FILE1_PATH) == "binary"


def test_guess_encoding_ascii():
    assert File.guess_encoding(TEST_ASCII_PATH) == "us-ascii"


def test_guess_encoding_unicode():
    assert File.guess_encoding(TEST_UNICODE_PATH) == "utf-8"


def test_guess_encoding_iso8859():
    assert File.guess_encoding(TEST_ISO8859_PATH) == "iso-8859-1"


def test_no_differences_with_xxd(binary1):
    difference = binary1.compare_bytes(binary1)
    assert difference is None


@skip_unless_tools_exist("xxd")
def test_compare_with_xxd(binary1, binary2):
    difference = binary1.compare_bytes(binary2)
    expected_diff = get_data("binary_expected_diff")
    assert normalize_zeros(difference.unified_diff) == expected_diff


def test_compare_non_existing_with_xxd(binary1):
    difference = binary1.compare_bytes(MissingFile("/nonexisting", binary1))
    assert difference.source2 == "/nonexisting"


@pytest.fixture
def xxd_not_found(monkeypatch):
    def mock_cmdline(self):
        raise RequiredToolNotFound("xxd")

    monkeypatch.setattr(Xxd, "cmdline", mock_cmdline)


def test_no_differences_without_xxd(xxd_not_found, binary1):
    difference = binary1.compare_bytes(binary1)
    assert difference is None


def test_compare_without_xxd(xxd_not_found, binary1, binary2):
    difference = binary1.compare(binary2)
    expected_diff = get_data("binary_hexdump_expected_diff")
    assert difference.unified_diff == expected_diff


def test_with_compare_details():
    d = Difference(TEST_FILE1_PATH, TEST_FILE2_PATH, source="source")

    class MockFile(FilesystemFile):
        def compare_details(self, other, source=None):
            return [d]

    difference = MockFile(TEST_FILE1_PATH).compare(
        MockFile(TEST_FILE2_PATH), source="source"
    )
    assert difference.details[0] == d


@skip_unless_tools_exist("xxd")
def test_with_compare_details_and_fallback():
    class MockFile(FilesystemFile):
        DESCRIPTION = "mock files"

        def compare_details(self, other, source=None):
            return []

    difference = MockFile(TEST_FILE1_PATH).compare(MockFile(TEST_FILE2_PATH))
    expected_diff = get_data("binary_expected_diff")
    assert (
        "supported for mock files but no file-specific differences were detected"
        in difference.comment
    )
    assert normalize_zeros(difference.unified_diff) == expected_diff


def test_with_compare_details_and_no_actual_differences():
    class MockFile(FilesystemFile):
        def compare_details(self, other, source=None):
            return []

    difference = MockFile(TEST_FILE1_PATH).compare(MockFile(TEST_FILE1_PATH))
    assert difference is None


@skip_unless_tools_exist("xxd")
def test_with_compare_details_and_failed_process():
    output = "Free Jeremy Hammond"

    class MockFile(FilesystemFile):
        def compare_details(self, other, source=None):
            subprocess.check_call(["sh", "-c", 'echo "%s"; exit 42' % output])
            raise Exception("should not be run")

    difference = MockFile(TEST_FILE1_PATH).compare(MockFile(TEST_FILE2_PATH))
    expected_diff = get_data("../data/binary_expected_diff")
    assert output in difference.comment
    assert "42" in difference.comment
    assert normalize_zeros(difference.unified_diff) == expected_diff


@skip_unless_tools_exist("xxd")
def test_with_compare_details_and_parsing_error():
    from diffoscope.exc import OutputParsingError

    class MockFile(FilesystemFile):
        def compare_details(self, other, source=None):
            subprocess.check_output(["sh", "-c", "exit 0"])
            raise OutputParsingError("sh", self)

    difference = MockFile(TEST_FILE1_PATH).compare(MockFile(TEST_FILE2_PATH))
    expected_diff = get_data("../data/binary_expected_diff")
    assert "Error parsing output" in difference.comment
    assert normalize_zeros(difference.unified_diff) == expected_diff


@skip_unless_tools_exist("xxd")
def test_with_compare_details_and_extraction_error():
    from diffoscope.exc import ContainerExtractionError

    class MockFile(FilesystemFile):
        def compare_details(self, other, source=None):
            subprocess.check_output(["sh", "-c", "exit 0"])
            raise ContainerExtractionError(self.path, Exception())

    difference = MockFile(TEST_FILE1_PATH).compare(MockFile(TEST_FILE2_PATH))
    expected_diff = get_data("../data/binary_expected_diff")
    assert "Error extracting" in difference.comment
    assert normalize_zeros(difference.unified_diff) == expected_diff


@skip_unless_tools_exist("xxd")
@skip_unless_module_exists("distro")
def test_with_compare_details_and_tool_not_found(monkeypatch):
    from diffoscope.external_tools import EXTERNAL_TOOLS

    monkeypatch.setitem(
        EXTERNAL_TOOLS,
        "nonexistent",
        {
            "debian": "some-package",
            "arch": "some-package",
            "fedora": "some-package",
            "FreeBSD": "some-package",
        },
    )

    class MockFile(FilesystemFile):
        @tool_required("nonexistent")
        def compare_details(self, other, source=None):
            raise Exception("should not be run")

    difference = MockFile(TEST_FILE1_PATH).compare(MockFile(TEST_FILE2_PATH))
    expected_diff = get_data("binary_expected_diff")
    assert "nonexistent" in difference.comment
    assert "some-package" in difference.comment
    assert normalize_zeros(difference.unified_diff) == expected_diff


def test_compare_two_nonexisting_files():
    file1 = MissingFile("/nonexisting1")
    file2 = MissingFile("/nonexisting2")
    difference = file1.compare(file2)
    assert "non-existing" in difference.comment


def test_symlink_to_dir():
    # Create 2 directories, each containing sub-directory src and symbolic link dst-->src.
    with TemporaryDirectory() as basepath1:
        with TemporaryDirectory() as basepath2:
            src1path = os.path.join(basepath1, "src")
            dst1path = os.path.join(basepath1, "lnk")
            src2path = os.path.join(basepath2, "src")
            dst2path = os.path.join(basepath2, "lnk")
            mkdir(src1path)
            mkdir(src2path)
            symlink(src1path, dst1path)
            symlink(src2path, dst2path)

            # Compare these directories' content.
            file1 = FilesystemFile(basepath1)
            file2 = FilesystemFile(basepath2)
            assert file1.has_same_content_as(file2) is False



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2015 Jérémy Bobbio <lunar@debian.org>
# Copyright © 2015-2020 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import pytest

from diffoscope.comparators.zip import ZipFile, MozillaZipFile, JmodJavaModule

from ..utils.data import load_fixture, get_data
from ..utils.tools import skip_unless_tools_exist
from ..utils.nonexisting import assert_non_existing


zip1 = load_fixture("test1.zip")
zip2 = load_fixture("test2.zip")
zip3 = load_fixture("test3.zip")
encrypted_zip1 = load_fixture("encrypted1.zip")
encrypted_zip2 = load_fixture("encrypted2.zip")
mozzip1 = load_fixture("test1.mozzip")
mozzip2 = load_fixture("test2.mozzip")
jmod1 = load_fixture("test1.jmod")
jmod2 = load_fixture("test2.jmod")
test_comment1 = load_fixture("test_comment1.zip")
test_comment2 = load_fixture("test_comment2.zip")


def test_identification(zip1):
    assert isinstance(zip1, ZipFile)


def test_no_differences(zip1):
    difference = zip1.compare(zip1)
    assert difference is None


@pytest.fixture
def differences(zip1, zip2):
    return zip1.compare(zip2).details


@pytest.fixture
def differences2(zip1, zip3):
    return zip1.compare(zip3).details


@skip_unless_tools_exist("zipinfo")
def test_metadata(differences):
    expected_diff = get_data("zip_zipinfo_expected_diff")
    assert differences[0].unified_diff == expected_diff


@skip_unless_tools_exist("zipinfo")
def test_compressed_files(differences):
    assert differences[1].source1 == "dir/text"
    assert differences[1].source2 == "dir/text"
    expected_diff = get_data("text_ascii_expected_diff")
    assert differences[1].unified_diff == expected_diff


@skip_unless_tools_exist("zipinfo", "bsdtar")
def test_extra_fields(differences2):
    expected_diff = get_data("zip_bsdtar_expected_diff")
    assert differences2[0].unified_diff == expected_diff


@skip_unless_tools_exist("zipinfo")
def test_compare_non_existing(monkeypatch, zip1):
    assert_non_existing(monkeypatch, zip1)


def test_mozzip_identification(mozzip1):
    assert isinstance(mozzip1, MozillaZipFile)


def test_mozzip_no_differences(mozzip1):
    difference = mozzip1.compare(mozzip1)
    assert difference is None


@pytest.fixture
def mozzip_differences(mozzip1, mozzip2):
    return mozzip1.compare(mozzip2).details


@skip_unless_tools_exist("zipinfo")
def test_mozzip_metadata(mozzip_differences, mozzip1, mozzip2):
    expected_diff = get_data("mozzip_zipinfo_expected_diff")
    assert mozzip_differences[0].unified_diff == expected_diff


@skip_unless_tools_exist("zipinfo")
def test_mozzip_compressed_files(mozzip_differences):
    assert mozzip_differences[-1].source1 == "dir/text"
    assert mozzip_differences[-1].source2 == "dir/text"
    expected_diff = get_data("text_ascii_expected_diff")
    assert mozzip_differences[-1].unified_diff == expected_diff


@skip_unless_tools_exist("zipinfo")
def test_mozzip_compare_non_existing(monkeypatch, mozzip1):
    assert_non_existing(monkeypatch, mozzip1)


def test_jmod_identification(jmod1):
    assert isinstance(jmod1, JmodJavaModule)


def test_jmod_no_differences(jmod1):
    difference = jmod1.compare(jmod1)
    assert difference is None


@pytest.fixture
def jmod_differences(jmod1, jmod2):
    return jmod1.compare(jmod2).details


@skip_unless_tools_exist("zipinfo")
def test_jmod_metadata(jmod_differences, jmod1, jmod2):
    expected_diff = get_data("jmod_zipinfo_expected_diff")
    assert jmod_differences[0].unified_diff == expected_diff


def test_encrypted(encrypted_zip1, encrypted_zip2):
    difference = encrypted_zip1.compare(encrypted_zip2)
    assert difference is not None


@pytest.fixture
def comment_differences(test_comment1, test_comment2):
    return test_comment1.compare(test_comment2).details


@skip_unless_tools_exist("zipnote")
def test_commented(comment_differences):
    expected_diff = get_data("comment_zipinfo_expected_diff")
    assert comment_differences[1].unified_diff == expected_diff



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2014-2015 Jérémy Bobbio <lunar@debian.org>
# Copyright © 2016 Ximin Luo <infinity0@pwned.gg>
# Copyright © 2016-2020 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import re
import logging

from diffoscope.tools import tool_required
from diffoscope.difference import Difference

from .utils.file import File
from .utils.command import Command
from .utils.libarchive import LibarchiveContainer, list_libarchive

logger = logging.getLogger(__name__)


# For Go archives this gives a readable plain diff of the __.PKGDEF
# member which is just a text file containing the Go interface.
# TODO: add Go tests and more recursion.


class ArContainer(LibarchiveContainer):
    def get_adjusted_members(self):
        members = list(super().get_adjusted_members())
        known_ignores = {
            "/": "this is the symbol table, already accounted for in other output",
            "//": "this is the table for GNU long names, already accounted for in the archive filelist",
        }
        filtered_out = [p for p in members if p[0] in known_ignores]
        if filtered_out:
            for k, v in filtered_out:
                logger.debug(
                    "ignored ar member '%s' because %s", k, known_ignores[k]
                )
        return [p for p in members if p[0] not in known_ignores]


class ArSymbolTableDumper(Command):
    @tool_required("nm")
    def cmdline(self):
        return ["nm", "-s", self.path]


class ArFile(File):
    DESCRIPTION = "ar(1) archives"
    CONTAINER_CLASSES = [ArContainer]
    FILE_TYPE_RE = re.compile(r"\bar archive\b")

    def compare_details(self, other, source=None):
        return [
            Difference.from_operation(
                ArSymbolTableDumper, self.path, other.path
            ),
            Difference.from_text_readers(
                list_libarchive(self.path),
                list_libarchive(other.path),
                self.path,
                other.path,
                source="file list",
            ),
        ]



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2018 Xavier Briand <xavierbriand@gmail.com>
# Copyright © 2020 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import shutil
import pytest

from diffoscope.comparators.lz4 import Lz4File
from diffoscope.comparators.binary import FilesystemFile
from diffoscope.comparators.utils.specialize import specialize

from ..utils.data import load_fixture, get_data
from ..utils.tools import skip_unless_tools_exist
from ..utils.nonexisting import assert_non_existing

lz41 = load_fixture("test1.lz4")
lz42 = load_fixture("test2.lz4")


def test_identification(lz41):
    assert isinstance(lz41, Lz4File)


def test_no_differences(lz41):
    difference = lz41.compare(lz41)
    assert difference is None


@pytest.fixture
def differences(lz41, lz42):
    return lz41.compare(lz42).details


@skip_unless_tools_exist("lz4")
def test_content_source(differences):
    assert differences[0].source1 == "test1"
    assert differences[0].source2 == "test2"


@skip_unless_tools_exist("lz4")
def test_content_source_without_extension(tmpdir, lz41, lz42):
    path1 = str(tmpdir.join("test1"))
    path2 = str(tmpdir.join("test2"))
    shutil.copy(lz41.path, path1)
    shutil.copy(lz42.path, path2)
    lz41 = specialize(FilesystemFile(path1))
    lz42 = specialize(FilesystemFile(path2))
    difference = lz41.compare(lz42).details
    assert difference[0].source1 == "test1-content"
    assert difference[0].source2 == "test2-content"


@skip_unless_tools_exist("lz4")
def test_content_diff(differences):
    expected_diff = get_data("text_ascii_expected_diff")
    assert differences[0].unified_diff == expected_diff


@skip_unless_tools_exist("lz4")
def test_compare_non_existing(monkeypatch, lz41):
    assert_non_existing(monkeypatch, lz41)



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2015      Jérémy Bobbio <lunar@debian.org>
# Copyright © 2016-2017 Mattia Rizzolo <mattia@debian.org>
# Copyright © 2018-2020 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

from diffoscope.config import Config
from diffoscope.comparators.missing_file import MissingFile


def assert_non_existing(
    monkeypatch, fixture, has_null_source=True, has_details=True
):
    monkeypatch.setattr(Config(), "new_file", True)
    assert Config().new_file, "Did not get patched"

    difference = fixture.compare(MissingFile("/nonexisting", fixture))

    assert difference.source2 == "/nonexisting"
    assert not has_details or len(difference.details) > 0
    assert not has_null_source or (
        difference.details[-1].source2 == "/dev/null"
        or difference.source2 == "/nonexisting"
    )



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2015 Jérémy Bobbio <lunar@debian.org>
# Copyright © 2015-2018, 2020 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import pytest
import subprocess

from diffoscope.config import Config
from diffoscope.comparators.java import ClassFile, ProcyonDecompiler, Javap
from diffoscope.comparators.missing_file import MissingFile

from ..utils.data import load_fixture, get_data
from ..utils.tools import (
    skip_unless_tools_exist,
    skip_unless_tool_is_at_least,
    skip_unless_tool_is_between,
)


class1 = load_fixture("Test1.class")
class2 = load_fixture("Test2.class")


def javap_version():
    try:
        out = subprocess.check_output(["javap", "-version"])
    except subprocess.CalledProcessError as e:
        out = e.output
    return out.decode("UTF-8").strip()


def test_identification(class1):
    assert isinstance(class1, ClassFile)


def test_no_differences(class1):
    difference = class1.compare(class1)
    assert difference is None


@pytest.fixture
def differences_procyon(monkeypatch, class1, class2):
    monkeypatch.setattr(class1, "decompilers", [ProcyonDecompiler])
    return class1.compare(class2).details


@pytest.fixture
def differences_javap(monkeypatch, class1, class2):
    monkeypatch.setattr(class1, "decompilers", [Javap])
    return class1.compare(class2).details


def diff(differences, expected_diff_file):
    expected_diff = get_data(expected_diff_file)
    assert differences[0].unified_diff == expected_diff


def compare_non_existing(monkeypatch, class1, decompiler):
    monkeypatch.setattr(Config(), "new_file", True)
    monkeypatch.setattr(class1, "decompilers", [decompiler])
    difference = class1.compare(MissingFile("/nonexisting", class1))
    assert difference.source2 == "/nonexisting"
    assert len(difference.details) > 0


@skip_unless_tools_exist("procyon")
def test_diff_procyon(differences_procyon):
    diff(differences_procyon, "procyon_class_expected_diff")


@skip_unless_tool_is_between("javap", javap_version, "9.0.4", "14.0")
def test_diff_javap(differences_javap):
    """
    Java tests are kept up to date relative to the JDK version bundled in
    Debian's `default-jdk` package. The output may vary depending on the JDK
    version installed on your system, rendering the tests inaccurate.
    """

    diff(differences_javap, "javap_class_expected_diff")


@skip_unless_tool_is_at_least("javap", javap_version, "14.0")
def test_diff_javap_14(differences_javap):
    diff(differences_javap, "javap_14_class_expected_diff")


@skip_unless_tools_exist("procyon")
def test_compare_non_existing_procyon(monkeypatch, class1):
    compare_non_existing(monkeypatch, class1, ProcyonDecompiler)


@skip_unless_tools_exist("javap")
def test_compare_non_existing_javap(monkeypatch, class1):
    compare_non_existing(monkeypatch, class1, Javap)



#
# diffoscope: in-depth comparison of files, archives, and directories
#
# Copyright © 2017, 2019-2020 Chris Lamb <lamby@debian.org>
#
# diffoscope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# diffoscope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with diffoscope.  If not, see <https://www.gnu.org/licenses/>.

import pytest

from diffoscope.comparators.image import ICOImageFile
from diffoscope.config import Config

from ..utils.data import load_fixture, get_data
from ..utils.tools import (
    skip_unless_tools_exist,
    skip_unless_tool_is_at_least,
)
from .test_jpeg_image import identify_version


image1 = load_fixture("test1.ico")
image2 = load_fixture("test2.ico")
image1_meta = load_fixture("test1_meta.ico")
image2_meta = load_fixture("test2_meta.ico")


def test_identification(image1):
    assert isinstance(image1, ICOImageFile)


def test_no_differences(image1):
    difference = image1.compare(image1)
    assert difference is None


@pytest.fixture
def differences(image1, image2):
    return image1.compare(image2).details


@skip_unless_tools_exist("img2txt", "convert")
def test_diff(differences):
    expected_diff = get_data("ico_image_expected_diff")
    assert differences[0].unified_diff == expected_diff


@pytest.fixture
def differences_meta(image1_meta, image2_meta):
    return image1_meta.compare(image2_meta).details


@skip_unless_tools_exist("img2txt", "identify")
@skip_unless_tool_is_at_least("identify", identify_version, "6.9.10-23")
def test_diff_meta(differences_meta):
    expected_diff = get_data("ico_image_meta_expected_diff")
    assert differences_meta[-1].unified_diff == expected_diff


@skip_unless_tools_exist("img2txt", "identify")
@skip_unless_tool_is_at_least("identify", identify_version, "6.9.10-23")
def test_diff_meta2(differences_meta):
    expected_diff = get_data("ico_image_meta_expected_diff_v2")
    assert differences_meta[-1].unified_diff == expected_diff


@skip_unless_tools_exist("img2txt", "convert", "identify")
def test_has_visuals(monkeypatch, image1, image2):
    monkeypatch.setattr(Config(), "compute_visual_diffs", True)
    ico_diff = image1.compare(image2)
    assert len(ico_diff.details) == 2
    assert len(ico_diff.details[0].visuals) == 2
    assert ico_diff.details[0].visuals[0].data_type == "image/png;base64"
    assert ico_diff.details[0].visuals[1].data_type == "image/gif;base64"
